"""
Stopwords in different languages.
"""

FRENCH_STOPWORDS = {
    "alors",
    "au",
    "aucuns",
    "aussi",
    "autre",
    "aux",
    "avant",
    "avec",
    "avoir",
    "bon",
    "car",
    "ce",
    "cela",
    "ces",
    "ceux",
    "chaque",
    "ci",
    "comme",
    "comment",
    "dans",
    "de",
    "dedans",
    "dehors",
    "depuis",
    "des",
    "deux",
    "devrait",
    "doit",
    "donc",
    "dos",
    "droite",
    "du",
    "début",
    "elle",
    "elles",
    "en",
    "encore",
    "essai",
    "est",
    "et",
    "eu",
    "eux",
    "fait",
    "faites",
    "fois",
    "font",
    "force",
    "haut",
    "hors",
    "ici",
    "il",
    "ils",
    "je",
    "juste",
    "la",
    "le",
    "les",
    "leur",
    "lui",
    "là",
    "ma",
    "maintenant",
    "mais",
    "me",
    "meme",
    "mes",
    "mine",
    "moi",
    "moins",
    "mon",
    "mot",
    "ne",
    "ni",
    "nommés",
    "nos",
    "notre",
    "nous",
    "nouveaux",
    "on",
    "ou",
    "où",
    "par",
    "parce",
    "parole",
    "pas",
    "personnes",
    "peu",
    "peut",
    "pièce",
    "plupart",
    "pour",
    "pourquoi",
    "qu",
    "quand",
    "que",
    "quel",
    "quelle",
    "quelles",
    "quels",
    "qui",
    "sa",
    "sans",
    "se",
    "ses",
    "seulement",
    "si",
    "sien",
    "son",
    "sont",
    "sous",
    "soyez",
    "sujet",
    "sur",
    "ta",
    "tandis",
    "te",
    "tellement",
    "tels",
    "tes",
    "toi",
    "ton",
    "tous",
    "tout",
    "trop",
    "très",
    "tu",
    "un",
    "une",
    "valeur",
    "voie",
    "voient",
    "vont",
    "vos",
    "votre",
    "vous",
    "vu",
    "ça",
    "étaient",
    "état",
    "étions",
    "été",
    "être",
}


ENGLISH_STOPWORDS = {
    "a",
    "able",
    "about",
    "above",
    "according",
    "accordingly",
    "across",
    "actually",
    "after",
    "afterwards",
    "again",
    "against",
    "ain't",
    "all",
    "allow",
    "allows",
    "almost",
    "alone",
    "along",
    "already",
    "also",
    "although",
    "always",
    "am",
    "among",
    "amongst",
    "amoungst",
    "amount",
    "an",
    "and",
    "another",
    "any",
    "anybody",
    "anyhow",
    "anyone",
    "anything",
    "anyway",
    "anyways",
    "anywhere",
    "apart",
    "appear",
    "appreciate",
    "appropriate",
    "are",
    "aren't",
    "around",
    "as",
    "aside",
    "ask",
    "asking",
    "associated",
    "at",
    "available",
    "away",
    "awfully",
    "back",
    "be",
    "became",
    "because",
    "become",
    "becomes",
    "becoming",
    "been",
    "before",
    "beforehand",
    "behind",
    "being",
    "believe",
    "below",
    "beside",
    "besides",
    "best",
    "better",
    "between",
    "beyond",
    "bill",
    "both",
    "bottom",
    "brief",
    "but",
    "by",
    "call",
    "came",
    "can",
    "cannot",
    "cant",
    "can't",
    "cause",
    "causes",
    "certain",
    "certainly",
    "changes",
    "clearly",
    "co",
    "com",
    "come",
    "comes",
    "computer",
    "con",
    "concerning",
    "consequently",
    "consider",
    "considering",
    "contain",
    "containing",
    "contains",
    "corresponding",
    "could",
    "couldnt",
    "couldn't",
    "course",
    "cry",
    "currently",
    "c'mon",
    "c's",
    "de",
    "definitely",
    "describe",
    "described",
    "despite",
    "detail",
    "did",
    "didn't",
    "different",
    "do",
    "does",
    "doesn't",
    "doing",
    "done",
    "don't",
    "down",
    "downwards",
    "due",
    "during",
    "each",
    "edu",
    "eg",
    "eight",
    "either",
    "eleven",
    "else",
    "elsewhere",
    "empty",
    "enough",
    "entirely",
    "especially",
    "et",
    "etc",
    "even",
    "ever",
    "every",
    "everybody",
    "everyone",
    "everything",
    "everywhere",
    "ex",
    "exactly",
    "example",
    "except",
    "far",
    "few",
    "fifteen",
    "fifth",
    "fify",
    "fill",
    "find",
    "fire",
    "first",
    "five",
    "followed",
    "following",
    "follows",
    "for",
    "former",
    "formerly",
    "forth",
    "forty",
    "found",
    "four",
    "from",
    "front",
    "full",
    "further",
    "furthermore",
    "get",
    "gets",
    "getting",
    "give",
    "given",
    "gives",
    "go",
    "goes",
    "going",
    "gone",
    "got",
    "gotten",
    "greetings",
    "had",
    "hadn't",
    "happens",
    "hardly",
    "has",
    "hasnt",
    "hasn't",
    "have",
    "haven't",
    "having",
    "he",
    "hello",
    "help",
    "hence",
    "her",
    "here",
    "hereafter",
    "hereby",
    "herein",
    "hereupon",
    "here's",
    "hers",
    "herself",
    "he's",
    "hi",
    "him",
    "himself",
    "his",
    "hither",
    "hopefully",
    "how",
    "howbeit",
    "however",
    "hundred",
    "i",
    "i'd",
    "i'll",
    "i'm",
    "i've",
    "ie",
    "if",
    "ignored",
    "immediate",
    "in",
    "inasmuch",
    "inc",
    "indeed",
    "indicate",
    "indicated",
    "indicates",
    "inner",
    "insofar",
    "instead",
    "interest",
    "into",
    "inward",
    "is",
    "isn't",
    "it",
    "its",
    "itself",
    "it'd",
    "it'll",
    "it's'",
    "i'd",
    "i'll",
    "i'm",
    "i've",
    "just",
    "keep",
    "keeps",
    "kept",
    "know",
    "known",
    "knows",
    "last",
    "lately",
    "later",
    "latter",
    "latterly",
    "least",
    "less",
    "lest",
    "let",
    "let's",
    "like",
    "liked",
    "likely",
    "little",
    "look",
    "looking",
    "looks",
    "ltd",
    "made",
    "mainly",
    "many",
    "may",
    "maybe",
    "me",
    "mean",
    "meanwhile",
    "merely",
    "might",
    "mill",
    "mine",
    "more",
    "moreover",
    "most",
    "mostly",
    "move",
    "much",
    "must",
    "my",
    "myself",
    "name",
    "namely",
    "nd",
    "near",
    "nearly",
    "necessary",
    "need",
    "needs",
    "neither",
    "never",
    "nevertheless",
    "new",
    "next",
    "nine",
    "no",
    "nobody",
    "non",
    "none",
    "noone",
    "nor",
    "normally",
    "not",
    "nothing",
    "novel",
    "now",
    "nowhere",
    "obviously",
    "of",
    "off",
    "often",
    "oh",
    "ok",
    "okay",
    "old",
    "on",
    "once",
    "one",
    "ones",
    "only",
    "onto",
    "or",
    "other",
    "others",
    "otherwise",
    "ought",
    "our",
    "ours",
    "ourselves",
    "out",
    "outside",
    "over",
    "overall",
    "own",
    "part",
    "particular",
    "particularly",
    "per",
    "perhaps",
    "placed",
    "please",
    "plus",
    "possible",
    "presumably",
    "probably",
    "provides",
    "put",
    "que",
    "quite",
    "qv",
    "rather",
    "rd",
    "re",
    "really",
    "reasonably",
    "regarding",
    "regardless",
    "regards",
    "relatively",
    "respectively",
    "right",
    "said",
    "same",
    "saw",
    "say",
    "saying",
    "says",
    "second",
    "secondly",
    "see",
    "seeing",
    "seem",
    "seemed",
    "seeming",
    "seems",
    "seen",
    "self",
    "selves",
    "sensible",
    "sent",
    "serious",
    "seriously",
    "seven",
    "several",
    "shall",
    "she",
    "should",
    "shouldn't",
    "show",
    "side",
    "since",
    "sincere",
    "six",
    "sixty",
    "so",
    "some",
    "somebody",
    "somehow",
    "someone",
    "something",
    "sometime",
    "sometimes",
    "somewhat",
    "somewhere",
    "soon",
    "sorry",
    "specified",
    "specify",
    "specifying",
    "still",
    "sub",
    "such",
    "sup",
    "sure",
    "system",
    "take",
    "taken",
    "tell",
    "ten",
    "tends",
    "th",
    "than",
    "thank",
    "thanks",
    "thanx",
    "that",
    "that's",
    "thats",
    "that's",
    "the",
    "their",
    "theirs",
    "them",
    "themselves",
    "then",
    "thence",
    "there",
    "thereafter",
    "thereby",
    "therefore",
    "therein",
    "theres",
    "thereupon",
    "there's",
    "these",
    "they",
    "they'd",
    "they'll",
    "they're",
    "they've",
    "thick",
    "thin",
    "think",
    "third",
    "this",
    "thorough",
    "thoroughly",
    "those",
    "though",
    "three",
    "through",
    "throughout",
    "thru",
    "thus",
    "to",
    "together",
    "too",
    "took",
    "top",
    "toward",
    "towards",
    "tried",
    "tries",
    "truly",
    "try",
    "trying",
    "twelve",
    "twenty",
    "twice",
    "two",
    "t's",
    "un",
    "under",
    "unfortunately",
    "unless",
    "unlikely",
    "until",
    "unto",
    "up",
    "upon",
    "us",
    "use",
    "used",
    "useful",
    "uses",
    "using",
    "usually",
    "value",
    "various",
    "very",
    "via",
    "viz",
    "vs",
    "want",
    "wants",
    "was",
    "wasn't",
    "way",
    "we",
    "welcome",
    "well",
    "went",
    "were",
    "weren't",
    "we'd",
    "we'll",
    "we're",
    "we've",
    "what",
    "whatever",
    "what's",
    "when",
    "whence",
    "whenever",
    "where",
    "whereafter",
    "whereas",
    "whereby",
    "wherein",
    "whereupon",
    "wherever",
    "where's",
    "whether",
    "which",
    "while",
    "whither",
    "who",
    "whoever",
    "whole",
    "whom",
    "whose",
    "who's",
    "why",
    "will",
    "willing",
    "wish",
    "with",
    "within",
    "without",
    "wonder",
    "won't",
    "would",
    "wouldn't",
    "yes",
    "yet",
    "you",
    "your",
    "yours",
    "yourself",
    "yourselves",
    "you'd",
    "you'll",
    "you're",
    "you've",
    "zero",
}


ENGLISH_REGULAR_VERBS = {
    "accept",
    "add",
    "admire",
    "admit",
    "advise",
    "afford",
    "agree",
    "alert",
    "allow",
    "amuse",
    "analyse",
    "announce",
    "annoy",
    "answer",
    "apologise",
    "appear",
    "applaud",
    "appreciate",
    "approve",
    "argue",
    "arrange",
    "arrest",
    "arrive",
    "ask",
    "attach",
    "attack",
    "attempt",
    "attend",
    "attract",
    "avoid",
    "back",
    "bake",
    "balance",
    "ban",
    "bang",
    "bare",
    "bat",
    "bathe",
    "battle",
    "beam",
    "beg",
    "behave",
    "belong",
    "bleach",
    "bless",
    "blind",
    "blink",
    "blot",
    "blush",
    "boast",
    "boil",
    "bolt",
    "bomb",
    "book",
    "bore",
    "borrow",
    "bounce",
    "bow",
    "box",
    "brake",
    "branch",
    "breathe",
    "bruise",
    "brush",
    "bubble",
    "bump",
    "burn",
    "bury",
    "buzz",
    "calculate",
    "call",
    "camp",
    "care",
    "carry",
    "carve",
    "cause",
    "challenge",
    "change",
    "charge",
    "chase",
    "cheat",
    "check",
    "cheer",
    "chew",
    "choke",
    "chop",
    "claim",
    "clap",
    "clean",
    "clear",
    "clip",
    "close",
    "coach",
    "coil",
    "collect",
    "colour",
    "comb",
    "command",
    "communicate",
    "compare",
    "compete",
    "complain",
    "complete",
    "concentrate",
    "concern",
    "confess",
    "confuse",
    "connect",
    "consider",
    "consist",
    "contain",
    "continue",
    "copy",
    "correct",
    "cough",
    "count",
    "cover",
    "crack",
    "crash",
    "crawl",
    "cross",
    "crush",
    "cry",
    "cure",
    "curl",
    "curve",
    "cycle",
    "dam",
    "damage",
    "dance",
    "dare",
    "decay",
    "deceive",
    "decide",
    "decorate",
    "delay",
    "delight",
    "deliver",
    "depend",
    "describe",
    "desert",
    "deserve",
    "destroy",
    "detect",
    "develop",
    "disagree",
    "disappear",
    "disapprove",
    "disarm",
    "discover",
    "dislike",
    "divide",
    "double",
    "doubt",
    "drag",
    "drain",
    "dream",
    "dress",
    "drip",
    "drop",
    "drown",
    "drum",
    "dry",
    "dust",
    "earn",
    "educate",
    "embarrass",
    "employ",
    "empty",
    "encourage",
    "end",
    "enjoy",
    "enter",
    "entertain",
    "escape",
    "examine",
    "excite",
    "excuse",
    "exercise",
    "exist",
    "expand",
    "expect",
    "explain",
    "explode",
    "extend",
    "face",
    "fade",
    "fail",
    "fancy",
    "fasten",
    "fax",
    "fear",
    "fence",
    "fetch",
    "file",
    "fill",
    "film",
    "fire",
    "fit",
    "fix",
    "flap",
    "flash",
    "float",
    "flood",
    "flow",
    "flower",
    "fold",
    "follow",
    "fool",
    "force",
    "form",
    "found",
    "frame",
    "frighten",
    "fry",
    "gather",
    "gaze",
    "glow",
    "glue",
    "grab",
    "grate",
    "grease",
    "greet",
    "grin",
    "grip",
    "groan",
    "guarantee",
    "guard",
    "guess",
    "guide",
    "hammer",
    "hand",
    "handle",
    "hang",
    "happen",
    "harass",
    "harm",
    "hate",
    "haunt",
    "head",
    "heal",
    "heap",
    "heat",
    "help",
    "hook",
    "hop",
    "hope",
    "hover",
    "hug",
    "hum",
    "hunt",
    "hurry",
    "identify",
    "ignore",
    "imagine",
    "impress",
    "improve",
    "include",
    "increase",
    "influence",
    "inform",
    "inject",
    "injure",
    "instruct",
    "intend",
    "interest",
    "interfere",
    "interrupt",
    "introduce",
    "invent",
    "invite",
    "irritate",
    "itch",
    "jail",
    "jam",
    "jog",
    "join",
    "joke",
    "judge",
    "juggle",
    "jump",
    "kick",
    "kill",
    "kiss",
    "kneel",
    "knit",
    "knock",
    "knot",
    "label",
    "land",
    "last",
    "laugh",
    "launch",
    "learn",
    "level",
    "license",
    "lick",
    "lie",
    "lighten",
    "like",
    "list",
    "listen",
    "live",
    "load",
    "lock",
    "long",
    "look",
    "love",
    "man",
    "manage",
    "march",
    "mark",
    "marry",
    "match",
    "mate",
    "matter",
    "measure",
    "meddle",
    "melt",
    "memorise",
    "mend",
    "mess up",
    "milk",
    "mine",
    "miss",
    "mix",
    "moan",
    "moor",
    "mourn",
    "move",
    "muddle",
    "mug",
    "multiply",
    "murder",
    "nail",
    "name",
    "need",
    "nest",
    "nod",
    "note",
    "notice",
    "number",
    "obey",
    "object",
    "observe",
    "obtain",
    "occur",
    "offend",
    "offer",
    "open",
    "order",
    "overflow",
    "owe",
    "own",
    "pack",
    "paddle",
    "paint",
    "park",
    "part",
    "pass",
    "paste",
    "pat",
    "pause",
    "peck",
    "pedal",
    "peel",
    "peep",
    "perform",
    "permit",
    "phone",
    "pick",
    "pinch",
    "pine",
    "place",
    "plan",
    "plant",
    "play",
    "please",
    "plug",
    "point",
    "poke",
    "polish",
    "pop",
    "possess",
    "post",
    "pour",
    "practise",
    "pray",
    "preach",
    "precede",
    "prefer",
    "prepare",
    "present",
    "preserve",
    "press",
    "pretend",
    "prevent",
    "prick",
    "print",
    "produce",
    "program",
    "promise",
    "protect",
    "provide",
    "pull",
    "pump",
    "punch",
    "puncture",
    "punish",
    "push",
    "question",
    "queue",
    "race",
    "radiate",
    "rain",
    "raise",
    "reach",
    "realise",
    "receive",
    "recognise",
    "record",
    "reduce",
    "reflect",
    "refuse",
    "regret",
    "reign",
    "reject",
    "rejoice",
    "relax",
    "release",
    "rely",
    "remain",
    "remember",
    "remind",
    "remove",
    "repair",
    "repeat",
    "replace",
    "reply",
    "report",
    "reproduce",
    "request",
    "rescue",
    "retire",
    "return",
    "rhyme",
    "rinse",
    "risk",
    "rob",
    "rock",
    "roll",
    "rot",
    "rub",
    "ruin",
    "rule",
    "rush",
    "sack",
    "sail",
    "satisfy",
    "save",
    "saw",
    "scare",
    "scatter",
    "scold",
    "scorch",
    "scrape",
    "scratch",
    "scream",
    "screw",
    "scribble",
    "scrub",
    "seal",
    "search",
    "separate",
    "serve",
    "settle",
    "shade",
    "share",
    "shave",
    "shelter",
    "shiver",
    "shock",
    "shop",
    "shrug",
    "sigh",
    "sign",
    "signal",
    "sin",
    "sip",
    "ski",
    "skip",
    "slap",
    "slip",
    "slow",
    "smash",
    "smell",
    "smile",
    "smoke",
    "snatch",
    "sneeze",
    "sniff",
    "snore",
    "snow",
    "soak",
    "soothe",
    "sound",
    "spare",
    "spark",
    "sparkle",
    "spell",
    "spill",
    "spoil",
    "spot",
    "spray",
    "sprout",
    "squash",
    "squeak",
    "squeal",
    "squeeze",
    "stain",
    "stamp",
    "stare",
    "start",
    "stay",
    "steer",
    "step",
    "stir",
    "stitch",
    "stop",
    "store",
    "strap",
    "strengthen",
    "stretch",
    "strip",
    "stroke",
    "stuff",
    "subtract",
    "succeed",
    "suck",
    "suffer",
    "suggest",
    "suit",
    "supply",
    "support",
    "suppose",
    "surprise",
    "surround",
    "suspect",
    "suspend",
    "switch",
    "talk",
    "tame",
    "tap",
    "taste",
    "tease",
    "telephone",
    "tempt",
    "terrify",
    "test",
    "thank",
    "thaw",
    "tick",
    "tickle",
    "tie",
    "time",
    "tip",
    "tire",
    "touch",
    "tour",
    "tow",
    "trace",
    "trade",
    "train",
    "transport",
    "trap",
    "travel",
    "treat",
    "tremble",
    "trick",
    "trip",
    "trot",
    "trouble",
    "trust",
    "try",
    "tug",
    "tumble",
    "turn",
    "twist",
    "type",
    "undress",
    "unfasten",
    "unite",
    "unlock",
    "unpack",
    "untidy",
    "use",
    "vanish",
    "visit",
    "wail",
    "wait",
    "walk",
    "wander",
    "want",
    "warm",
    "warn",
    "wash",
    "waste",
    "watch",
    "water",
    "wave",
    "weigh",
    "welcome",
    "whine",
    "whip",
    "whirl",
    "whisper",
    "whistle",
    "wink",
    "wipe",
    "wish",
    "wobble",
    "wonder",
    "work",
    "worry",
    "wrap",
    "wreck",
    "wrestle",
    "wriggle",
    "x-ray",
    "yawn",
    "yell",
    "zip",
    "zoom",
}


ENGLISH_IRREGULAR_VERBS = {
    "arise ",
    "arisen",
    "arose ",
    "ate",
    "awake",
    "awakened",
    "awoke",
    "awoken",
    "backslid",
    "backslidden",
    "backslide",
    "bade",
    "be",
    "bear",
    "beat",
    "beaten",
    "became",
    "become",
    "been",
    "began",
    "begin",
    "begun",
    "bend",
    "bent",
    "bet",
    "betted",
    "bid",
    "bidden",
    "bind",
    "bit",
    "bite",
    "bitten",
    "bled",
    "bleed",
    "blew",
    "blow",
    "blown",
    "bore",
    "born",
    "borne",
    "bought",
    "bound",
    "break",
    "bred",
    "breed",
    "bring",
    "broadcast",
    "broadcasted",
    "broke",
    "broken",
    "brought",
    "build",
    "built",
    "burn",
    "burned",
    "burnt",
    "burst",
    "bust",
    "busted",
    "buy",
    "came",
    "cast",
    "catch",
    "caught",
    "choose",
    "chose",
    "chosen",
    "clad",
    "cling",
    "clothe",
    "clothed",
    "clung",
    "come",
    "cost",
    "creep",
    "crept",
    "cut",
    "daydream",
    "daydreamed",
    "daydreamt",
    "deal",
    "dealt",
    "did",
    "dig",
    "disprove",
    "disproved",
    "disproven",
    "dive",
    "dived",
    "do",
    "done",
    "dove",
    "drank",
    "draw",
    "drawn",
    "dream",
    "dreamed",
    "dreamt",
    "drew",
    "drink",
    "drive",
    "driven",
    "drove",
    "drunk",
    "dug",
    "dwell",
    "dwelled",
    "dwelt",
    "eat",
    "eaten",
    "fall",
    "fallen",
    "fed",
    "feed",
    "feel",
    "fell",
    "felt",
    "fight",
    "find",
    "fit",
    "fitted",
    "fled",
    "flee",
    "flew",
    "fling",
    "flown",
    "flung",
    "fly",
    "forbade",
    "forbid",
    "forbidden",
    "forecast",
    "forego",
    "foregone",
    "foresaw",
    "foresee",
    "foreseen",
    "foretell",
    "foretold",
    "forewent",
    "forgave",
    "forget",
    "forgive",
    "forgiven",
    "forgot",
    "forgotten",
    "forsake",
    "forsaken",
    "forsook",
    "fought",
    "found",
    "freeze",
    "froze",
    "frozen",
    "gave",
    "get",
    "give",
    "given",
    "go",
    "gone",
    "got",
    "gotten",
    "grew",
    "grind",
    "ground",
    "grow",
    "grown",
    "had",
    "hang",
    "have",
    "hear",
    "heard",
    "held",
    "hew",
    "hewed",
    "hewn",
    "hid",
    "hidden",
    "hide",
    "hit",
    "hold",
    "hung",
    "hurt",
    "keep",
    "kept",
    "kneel",
    "kneeled",
    "knelt",
    "knew",
    "knit",
    "knitted",
    "know",
    "known",
    "laid",
    "lain",
    "lay",
    "lead",
    "lean",
    "leaned",
    "leant",
    "leap",
    "leaped",
    "leapt",
    "learn",
    "learned",
    "learnt",
    "leave",
    "led",
    "left",
    "lend",
    "lent",
    "let",
    "lie",
    "lied",
    "light",
    "lighted",
    "lit",
    "lose",
    "lost",
    "made",
    "make",
    "mean",
    "meant",
    "meet",
    "met",
    "misunderstand",
    "misunderstood",
    "mow",
    "mowed",
    "mown",
    "paid",
    "partake",
    "partaken",
    "partook",
    "pay",
    "plead",
    "pleaded",
    "pled",
    "proofread",
    "prove",
    "proved",
    "proven",
    "put",
    "quick-freeze",
    "quick-froze",
    "quick-frozen",
    "quit",
    "quitted",
    "ran",
    "rang",
    "read",
    "rid",
    "ridden",
    "ride",
    "ring",
    "rise",
    "risen",
    "rode",
    "rose",
    "run",
    "rung",
    "said",
    "sang",
    "sank",
    "sat",
    "saw",
    "sawed",
    "sawn",
    "say",
    "see",
    "seek",
    "seen",
    "sell",
    "send",
    "sent",
    "set",
    "sew",
    "sewed",
    "sewn",
    "shake",
    "shaken",
    "shave",
    "shaved",
    "shaven",
    "shear",
    "sheared",
    "shed",
    "shine",
    "shined",
    "shone",
    "shook",
    "shoot",
    "shorn",
    "shot",
    "show",
    "showed",
    "shown",
    "shrank",
    "shrink",
    "shrunk",
    "shut",
    "sing",
    "sink",
    "sit",
    "slain",
    "slay",
    "slayed",
    "sleep",
    "slept",
    "slew",
    "slid",
    "slide",
    "sling",
    "slink",
    "slinked",
    "slit",
    "slung",
    "slunk",
    "smell",
    "smelled",
    "smelt",
    "sneak",
    "sneaked",
    "snuck",
    "sold",
    "sought",
    "sow",
    "sowed",
    "sown",
    "spat",
    "speak",
    "sped",
    "speed",
    "speeded",
    "spell",
    "spelled",
    "spelt",
    "spend",
    "spent",
    "spill",
    "spilled",
    "spilt",
    "spin",
    "spit",
    "split",
    "spoil",
    "spoiled",
    "spoilt",
    "spoke",
    "spoken",
    "sprang",
    "spread",
    "spring",
    "sprung",
    "spun",
    "stand ",
    "stank",
    "steal",
    "stick",
    "sting",
    "stink",
    "stole",
    "stolen",
    "stood",
    "strew",
    "strewed",
    "strewn",
    "stricken",
    "stridden",
    "stride",
    "strike",
    "string",
    "strive",
    "strived",
    "striven",
    "strode",
    "strove",
    "struck",
    "strung",
    "stuck",
    "stung",
    "stunk",
    "sublet",
    "sunburn",
    "sunburned",
    "sunburnt",
    "sung",
    "sunk",
    "swam",
    "swear",
    "sweat",
    "sweated",
    "sweep",
    "swell",
    "swelled",
    "swept",
    "swim",
    "swing",
    "swollen",
    "swore",
    "sworn",
    "swum",
    "swung",
    "take",
    "taken",
    "taught",
    "teach",
    "tear",
    "telecast",
    "tell",
    "test-drive",
    "test-driven",
    "test-drove",
    "test-flew",
    "test-flown",
    "test-fly",
    "think",
    "thought",
    "threw",
    "throw",
    "thrown",
    "thrust",
    "told",
    "took",
    "tore",
    "torn",
    "tread",
    "trod",
    "trodden",
    "understand",
    "understood",
    "undertake",
    "undertaken",
    "undertook",
    "undid",
    "undo",
    "undone",
    "wake",
    "waked",
    "was, were",
    "waylaid",
    "waylay",
    "wear",
    "weave",
    "weaved",
    "wed",
    "wedded",
    "weep",
    "went",
    "wept",
    "wet",
    "wetted",
    "whet",
    "whetted",
    "win",
    "wind",
    "withdraw",
    "withdrawn",
    "withdrew",
    "withheld",
    "withhold",
    "withstand",
    "withstood",
    "woke",
    "woken",
    "won",
    "wore",
    "worn",
    "wound",
    "wove",
    "woven",
    "wring",
    "write",
    "written",
    "wrote",
    "wrung",
}



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import difflib
from functools import partial
from math import cos, sqrt, pi  # Needed for geographical distance

try:
    from dateutil import parser as dateparser

    DATEUTIL_ENABLED = True
except ImportError:
    DATEUTIL_ENABLED = False
from numpy import array, empty

from nazca.utils.normalize import tokenize

MAX_AUTOMATIC_CONFIDENCE = 0.9


###############################################################################
# UTILITY FUNCTIONS ###########################################################
###############################################################################
def distance_to_confidence(
    distance,
    max_confidence=MAX_AUTOMATIC_CONFIDENCE,
    max_distance=1,
):
    """Return a confidence between 0 and max_confidence based on the result of a Nazca
    processing
        :params distance: float (between 0 (exact match) and `max_distance`)
        :params max_confidence: the maximum confidence that can be returned.
        :params max_distance: the maximal distance that be used as input
            (usually 1, when a normalisation is used)
        :return: the confidence (float)
    """
    return min(max_confidence, round((max_distance - distance) / max_distance, 2))


def cdist(
    distance_callback,
    refset,
    targetset,
    matrix_normalized=False,
    ref_indexes=None,
    target_indexes=None,
):
    """Compute the metric matrix, given two datasets and a metric

    Parameters
    ----------
    refset: a dataset (list of records)

    targetset: a dataset (list of records)

    Returns
    -------

    A distance matrix, of shape (len(refset), len(targetset))
    with the distance of each element in it.
    """
    ref_indexes = ref_indexes or range(len(refset))
    target_indexes = target_indexes or range(len(targetset))
    distmatrix = empty((len(ref_indexes), len(target_indexes)), dtype="float32")
    for i, iref in enumerate(ref_indexes):
        for j, jref in enumerate(target_indexes):
            d = 1
            if refset[iref] and targetset[jref]:
                d = distance_callback(refset[iref], targetset[jref])
                if matrix_normalized:
                    d = 1 - (1.0 / (1.0 + d))
            distmatrix[i, j] = d
    return distmatrix


def _handlespaces(stra, strb, distance, tokenizer=None, **kwargs):
    """Compute the matrix of distances between all tokens of stra and strb
    (with function ``distance``). Extra args are given to the distance
    function

    The distance returned is defined as the max of the min of each rows of
    each distance matrix, see the example above :

             |  Victor |  Hugo                  Victor | Jean | Hugo
     Victor  |     0   |    5           Victor |  0    |  6   |  5
      Jean   |     6   |    4           Hugo   |  5    |  4   |  0
      Hugo   |     5   |    0

             --> 4                                --> 0

    Return 4
    """

    if " " not in stra:
        stra += " "
    if " " not in strb:
        strb += " "

    toka = tokenize(stra, tokenizer)
    tokb = tokenize(strb, tokenizer)
    # If not same number of tokens, complete the smallest list with empty
    # strings
    if len(toka) != len(tokb):
        mint = toka if len(toka) < len(tokb) else tokb
        maxt = toka if len(toka) > len(tokb) else tokb
        mint.extend(["" for i in range(len(maxt) - len(mint))])

    listmatrix = []
    for i in range(len(toka)):
        listmatrix.append([distance(toka[i], tokb[j], **kwargs) for j in range(len(tokb))])
    m = array(listmatrix)
    minlist = [m[i, :].min() for i in range(m.shape[0])]
    minlist.extend([m[:, i].min() for i in range(m.shape[1])])
    return max(minlist)


###############################################################################
# NUMERICAL DISTANCES #########################################################
###############################################################################
def euclidean(a, b):
    """Simple euclidian distance"""
    try:
        return abs(a - b)
    except TypeError:
        # a and b may be strings
        return abs(float(a) - float(b))


###############################################################################
# STRING DISTANCES ############################################################
###############################################################################
def exact_match(a, b):
    """The simplest distance, defined as 0 if both values are equal, 1 elsewise."""
    return 0 if a == b else 1


def levenshtein(stra, strb, tokenizer=None):
    """Compute the Levenshtein distance between stra and strb.

    The Levenshtein distance is defined as the minimal cost to transform stra
    into strb, where 3 operators are allowed :
        - Replace one character of stra into a character of strb
        - Add one character of strb into stra
        - Remove one character of strb

        If spaces are found in stra or strb, this method returns
            _handlespaces(stra, strb, levenshtein)
    """
    if " " in stra or " " in strb:
        return _handlespaces(stra, strb, levenshtein, tokenizer)

    lenb = len(strb)
    onerowago = None
    thisrow = list(range(1, lenb + 1)) + [0]
    for x in range(len(stra)):
        onerowago, thisrow = thisrow, [0] * lenb + [x + 1]
        for y in range(lenb):
            delcost = onerowago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = onerowago[y - 1] + (stra[x] != strb[y])
            thisrow[y] = min(delcost, addcost, subcost)
    return thisrow[lenb - 1]


def soundexcode(word, language="french"):
    """Return the Soundex code of the word ``word``
    For more information about soundex code see wiki_

    ``language`` can be 'french' or 'english'

    .:: wiki_ : https://en.wikipedia.org/wiki/Soundex

    If spaces are found in stra or strb, this method returns
        _handlespaces(stra, strb), soundex, language=language)
    """

    vowels = "AEHIOUWY"
    if language.lower() == "french":
        consonnantscode = {
            "B": "1",
            "P": "1",
            "C": "2",
            "K": "2",
            "Q": "2",
            "D": "3",
            "T": "3",
            "L": "4",
            "M": "5",
            "N": "5",
            "R": "6",
            "G": "7",
            "J": "7",
            "X": "8",
            "Z": "8",
            "S": "8",
            "F": "9",
            "V": "9",
        }
    elif language.lower() == "english":
        consonnantscode = {
            "B": "1",
            "F": "1",
            "P": "1",
            "V": "1",
            "C": "2",
            "G": "2",
            "J": "2",
            "K": "2",
            "Q": "2",
            "S": "2",
            "X": "2",
            "Z": "2",
            "D": "3",
            "T": "3",
            "L": "4",
            "M": "5",
            "N": "5",
            "R": "6",
        }
    else:
        raise NotImplementedError(
            "Soundex code is not supported (yet ?) for"
            "this language (%s). "
            "Supported languages are french and english" % language
        )
    word = word.strip().upper()
    code = word[0]
    # After this ``for`` code is
    # the first letter of ``word`` followed by all the consonnants of word,
    # where from consecutive consonnants, only the first is kept,
    # and from two identical consonnants separated by a W or a H, only the first
    # is kept too.
    for i in range(1, len(word)):
        if word[i] in vowels:
            continue
        if word[i - 1] not in vowels and consonnantscode[word[i]] == consonnantscode.get(
            code[-1], ""
        ):
            continue
        if (
            i + 2 < len(word)
            and word[i + 1] in "WH"
            and consonnantscode[word[i]] == consonnantscode.get(word[i + 2], "")
        ):
            continue
        code += word[i]
        if len(code) > 4:
            break

    # Replace according to the codes
    code = code[0] + "".join([consonnantscode[c] for c in code[1:]])
    # First four letters, completed by zeros
    return code[:4] + "0" * (4 - len(code))


def soundex(stra, strb, language="french", tokenizer=None):
    """Return the 1/0 distance between the soundex code of stra and strb.
    0 means they have the same code, 1 they don't
    """
    if " " in stra or " " in strb:
        return _handlespaces(stra, strb, soundex, tokenizer=tokenizer, language=language)
    return 0 if (soundexcode(stra, language) == soundexcode(strb, language)) else 1


def jaccard(stra, strb, tokenizer=None):
    r"""Return the jaccard distance between stra and strb, condering the tokens
    set of stra and strb. If no tokenizer is given, it use if
    alignement.normalize.tokenize's default one.

    J(A, B) = (A \cap B)/(A \cup B)
    d(A, B) = 1 - J(A, B)
    """
    seta = set(tokenize(stra, tokenizer))
    setb = set(tokenize(strb, tokenizer))
    return generic_jaccard(seta, setb)


def generic_jaccard(seta, setb):
    r"""Return the jaccard distance between two sets A and B.

    J(A, B) = (A \cap B)/(A \cup B)
    d(A, B) = 1 - J(A, B)

    special case: if A and B are emtpy, return 1.
    """
    try:
        return 1.0 - 1.0 * len(seta.intersection(setb)) / len(seta.union(setb))
    except ZeroDivisionError:
        return 1.0


def difflib_match(stra, strb):
    """Approximate matching.
    Extract of SequenceMatched documentation
    '[...] The basic algorithm predates, and is a little fancier than, an algorithm
    published in the late 1980's by Ratcliff and Obershelp under the
    hyperbolic name "gestalt pattern matching"[...]'

    A value smaller than 0.4 means that sequences are close matches (we take
    1 - difflib.SequenceMatched)
    """
    return 1.0 - difflib.SequenceMatcher(None, stra, strb).ratio()


###############################################################################
# TEMPORAL DISTANCES ##########################################################
###############################################################################
if DATEUTIL_ENABLED:

    class FrenchParserInfo(dateparser.parserinfo):
        """Inherit of the dateutil.parser.parserinfo and translate the english
        dependant variables into french.
        """

        HMS = [
            ("h", "heure", "heures"),
            ("m", "minute", "minutes"),
            ("s", "seconde", "seconde"),
        ]
        JUMP = [" ", ".", ",", ";", "-", "/", "'", "a", "le", "et", "er"]
        MONTHS = [
            ("Jan", "Janvier"),
            ("Fev", "Fevrier"),
            ("Mar", "Mars"),
            ("Avr", "Avril"),
            ("Mai", "Mai"),
            ("Jun", "Juin"),
            ("Jui", "Juillet"),
            ("Aou", "Aout"),
            ("Sep", "Septembre"),
            ("Oct", "Octobre"),
            ("Nov", "Novembre"),
            ("Dec", "Decembre"),
        ]
        PERTAIN = ["de"]
        WEEKDAYS = [
            ("Lun", "Lundi"),
            ("Mar", "Mardi"),
            ("Mer", "Mercredi"),
            ("Jeu", "Jeudi"),
            ("Ven", "Vendredi"),
            ("Sam", "Samedi"),
            ("Dim", "Dimanche"),
        ]

    def temporal(
        stra, strb, granularity="days", parserinfo=FrenchParserInfo, dayfirst=True, yearfirst=False
    ):
        """Return the distance between two strings (read as dates).

        ``granularity`` can be either ``days`` or ``months`` or ``years``
        (be careful to the plural form !)
        ``language`` can be either french or english

        ``dayfirst`` and ``yearfirst`` are used in case of ambiguity, for
        instance 09/09/09, by default it assumes it's day/month/year

        Neither stra nor strb can have accent. Clean it before.
        """

        datea = dateparser.parse(stra, parserinfo=parserinfo(dayfirst, yearfirst), fuzzy=True)
        dateb = dateparser.parse(strb, parserinfo=parserinfo(dayfirst, yearfirst), fuzzy=True)
        diff = datea - dateb
        if granularity.lower() == "years":
            return abs(diff.days / 365.25)
        if granularity.lower() == "months":
            return abs(diff.days / 30.5)
        return abs(diff.days)


###############################################################################
# GEOGRAPHICAL DISTANCES ######################################################
###############################################################################
def geographical(pointa, pointb, in_radians=False, planet_radius=6371009, units="m"):
    """Return the geographical distance between two points.

    Both points must be tuples (latitude, longitude)

    - in_radians is True, if latitude and longitude are in radians, false
      otherwise
    - planetRadius is the planet's radius in meters. By default, it's the
      Earth'one.

    - `units` can be 'm' (meters) or 'km' (kilometers)
    """
    if units not in ("m", "km"):
        raise ValueError("unsupported units, should be in m or km")
    pointa = (float(pointa[0]), float(pointa[1]))
    pointb = (float(pointb[0]), float(pointb[1]))

    difflat = pointa[0] - pointb[0]
    difflong = pointa[1] - pointb[1]
    meanlat = (pointa[0] + pointb[0]) / 2.0

    if not in_radians:
        difflat *= pi / 180.0
        difflong *= pi / 180.0
        meanlat *= pi / 180.0

    coef = 1.0 if units == "m" else 0.001
    return coef * planet_radius * sqrt(difflat**2 + (cos(meanlat) * difflong) ** 2)


###############################################################################
# BASE PROCESSING ########################################################
###############################################################################
class BaseProcessing:
    """A processing object used to provide an abstraction over the different
    distance functions, and help building Nazca process."""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        distance_callback=euclidean,
        weight=1,
        matrix_normalized=False,
    ):
        """Initiate the BaseProcessing

        Parameters
        ----------

        ref_attr_index: index of the attribute of interest in a record
                        for the reference dataset
                        (i.e. attribute to be used for key computation)

        target_attr_index: index of the attribute of interest in a record
                           for the target dataset
                           (i.e. attribute to be used for key computation)

        distance_callback: distance callback. Default is euclidean distance.

        weight: weight of the processing in a global distance matrix

        matrix_normalized: Boolean. If matrix_normalized is True,
                           the distance between two points is changed to
                           a value between 0 (equal) and 1 (totaly different).
                           To avoid useless computation and scale
                           problems the following “normalization” is done:
                                d = 1 - 1/(1 + d(x, y))

        """
        self.ref_attr_index = ref_attr_index
        self.target_attr_index = target_attr_index
        self.distance_callback = distance_callback
        self.weight = weight
        self.matrix_normalized = matrix_normalized

    def build_record(self, record, index):
        """Allow to have ref_attr_index and target_attr_index to be couple
        of index for (latitude, longitude)"""
        if isinstance(index, tuple) and len(index) == 2:
            return (record[index[0]], record[index[1]])
        else:
            return record[index] if index is not None else record

    def distance(self, reference_record, target_record):
        """Compute the distance between two records

        Parameters
        ----------
        reference_record: a record (tuple/list of values) of the reference dataset.

        target_record: a record (tuple/list of values) of the target dataset.

        """
        return self.distance_callback(
            self.build_record(reference_record, self.ref_attr_index),
            self.build_record(target_record, self.target_attr_index),
        )

    def cdist(self, refset, targetset, ref_indexes=None, target_indexes=None):
        """Compute the metric matrix, given two datasets and a metric

        Parameters
        ----------
        refset: a dataset (list of records)

        targetset: a dataset (list of records)

        Returns
        -------

        A distance matrix, of shape (len(refset), len(targetset))
        with the distance of each element in it.
        """
        return cdist(
            self.distance,
            refset,
            targetset,
            matrix_normalized=self.matrix_normalized,
            ref_indexes=ref_indexes,
            target_indexes=target_indexes,
        )

    def pdist(self, dataset):
        """Compute the upper triangular matrix in a way similar
        to scipy.spatial.metric

        Parameters
        ----------
        dataset: a dataset (list of records)

        Returns
        -------

        The values of the upper triangular distance matrix
        (of shape (len(dataset), len(dataset)) with the distance of each element in it.
        The values are sorted as row 1, row2, ...
        """
        values = []
        for i in range(len(dataset)):
            for j in range(i + 1, len(dataset)):
                d = 1
                if dataset[i] and dataset[j]:
                    d = self.distance(dataset[i], dataset[j])
                    if self.matrix_normalized:
                        d = 1 - (1.0 / (1.0 + d))
                values.append(d)
        return values


###############################################################################
# CONCRETE PROCESSINGS ###################################################
###############################################################################
class ExactMatchProcessing(BaseProcessing):
    """A processing based on the exact match (1 if a==b, 0 elsewise)"""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        tokenizer=None,
        weight=1,
        matrix_normalized=False,
    ):
        super().__init__(ref_attr_index, target_attr_index, exact_match, weight, matrix_normalized)


class LevenshteinProcessing(BaseProcessing):
    """A processing based on the levenshtein distance."""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        tokenizer=None,
        weight=1,
        matrix_normalized=False,
    ):
        distance_callback = partial(levenshtein, tokenizer=tokenizer)
        super().__init__(
            ref_attr_index, target_attr_index, distance_callback, weight, matrix_normalized
        )


class GeographicalProcessing(BaseProcessing):
    """A processing based on the geographical distance."""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        in_radians=False,
        planet_radius=6371009,
        units="m",
        weight=1,
        matrix_normalized=False,
    ):
        distance_callback = partial(
            geographical, in_radians=in_radians, planet_radius=planet_radius, units=units
        )
        super().__init__(
            ref_attr_index, target_attr_index, distance_callback, weight, matrix_normalized
        )


class SoundexProcessing(BaseProcessing):
    """A processing based on the soundex distance."""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        tokenizer=None,
        weight=1,
        language="french",
        matrix_normalized=False,
    ):
        distance_callback = partial(soundex, language=language, tokenizer=tokenizer)
        super().__init__(
            ref_attr_index, target_attr_index, distance_callback, weight, matrix_normalized
        )


class JaccardProcessing(BaseProcessing):
    """A processing based on the jaccard distance."""

    def __init__(
        self,
        ref_attr_index=None,
        target_attr_index=None,
        tokenizer=None,
        weight=1,
        matrix_normalized=False,
    ):
        distance_callback = partial(jaccard, tokenizer=tokenizer)
        super().__init__(
            ref_attr_index, target_attr_index, distance_callback, weight, matrix_normalized
        )


class DifflibProcessing(BaseProcessing):
    """A processing based on the difflib distance."""

    def __init__(
        self, ref_attr_index=None, target_attr_index=None, weight=1, matrix_normalized=False
    ):
        super().__init__(
            ref_attr_index, target_attr_index, difflib_match, weight, matrix_normalized
        )


if DATEUTIL_ENABLED:

    class TemporalProcessing(BaseProcessing):
        """A processing based on the temporal distance."""

        def __init__(
            self,
            ref_attr_index=None,
            target_attr_index=None,
            granularity="days",
            parserinfo=FrenchParserInfo,
            dayfirst=True,
            yearfirst=False,
            weight=1,
            matrix_normalized=False,
        ):
            distance_callback = partial(
                temporal,
                granularity=granularity,
                parserinfo=parserinfo,
                dayfirst=dayfirst,
                yearfirst=yearfirst,
            )
            super().__init__(
                ref_attr_index, target_attr_index, distance_callback, weight, matrix_normalized
            )



#
# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest
from os import path
from functools import partial
import random

from nazca.utils.distances import soundexcode
from nazca.rl.blocking import (
    KeyBlocking,
    SortedNeighborhoodBlocking,
    MergeBlocking,
    NGramBlocking,
    PipelineBlocking,
    SoundexBlocking,
    KmeansBlocking,
    MinHashingBlocking,
    KdTreeBlocking,
)
from nazca.utils.normalize import SimplifyNormalizer
from nazca.data import FRENCH_LEMMAS

# Make sure tests are repeatable / Minhashing
random.seed(6)


TESTDIR = path.dirname(__file__)

SOUNDEX_REFSET = (
    ("a1", "smith"),
    ("a2", "neighan"),
    ("a3", "meier"),
    ("a4", "smithers"),
    ("a5", "nguyen"),
    ("a6", "faulkner"),
    ("a7", "sandy"),
)
SOUNDEX_TARGETSET = (
    ("b1", "meier"),
    ("b2", "meier"),
    ("b3", "smith"),
    ("b4", "nguyen"),
    ("b5", "fawkner"),
    ("b6", "santi"),
    ("b7", "cain"),
)
SOUNDEX_PAIRS = (
    ("a3", "b1"),
    ("a3", "b2"),
    ("a2", "b4"),
    ("a5", "b4"),
    ("a1", "b3"),
    ("a1", "b6"),
    ("a7", "b3"),
    ("a7", "b6"),
)


class BaseBlockingTest(unittest.TestCase):
    def test_baseblocking_blocks(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn(([(0, "a1"), (6, "a7")], [(2, "b3"), (5, "b6")]), blocks)
        self.assertIn(([(1, "a2"), (4, "a5")], [(3, "b4")]), blocks)
        self.assertIn(([(2, "a3")], [(0, "b1"), (1, "b2")]), blocks)

    def test_baseblocking_id_blocks(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn((["a1", "a7"], ["b3", "b6"]), blocks)
        self.assertIn((["a2", "a5"], ["b4"]), blocks)
        self.assertIn((["a3"], ["b1", "b2"]), blocks)

    def test_baseblocking_indice_blocks(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_indice_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn(([0, 6], [2, 5]), blocks)
        self.assertIn(([1, 4], [3]), blocks)
        self.assertIn(([2], [0, 1]), blocks)

    def test_baseblocking_pairs(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        pairs = list(blocking.iter_pairs())
        ref_ind = {r[0]: ind for ind, r in enumerate(SOUNDEX_REFSET)}
        target_ind = {r[0]: ind for ind, r in enumerate(SOUNDEX_TARGETSET)}
        true_pairs = [((ref_ind[r[0]], r[0]), (target_ind[r[1]], r[1])) for r in SOUNDEX_PAIRS]
        self.assertEqual(len(pairs), len(true_pairs))
        for pair in true_pairs:
            self.assertIn(pair, pairs)

    def test_baseblocking_id_pairs(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        pairs = list(blocking.iter_id_pairs())
        true_pairs = SOUNDEX_PAIRS
        self.assertEqual(len(pairs), len(true_pairs))
        for pair in true_pairs:
            self.assertIn(pair, pairs)

    def test_baseblocking_indice_pairs(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        pairs = list(blocking.iter_indice_pairs())
        ref_ind = {r[0]: ind for ind, r in enumerate(SOUNDEX_REFSET)}
        target_ind = {r[0]: ind for ind, r in enumerate(SOUNDEX_TARGETSET)}
        true_pairs = [(ref_ind[r[0]], target_ind[r[1]]) for r in SOUNDEX_PAIRS]
        self.assertEqual(len(pairs), len(true_pairs))
        for pair in true_pairs:
            self.assertIn(pair, pairs)


class KeyBlockingTest(unittest.TestCase):
    def test_keyblocking_blocks(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn((["a1", "a7"], ["b3", "b6"]), blocks)
        self.assertIn((["a2", "a5"], ["b4"]), blocks)
        self.assertIn((["a3"], ["b1", "b2"]), blocks)

    def test_keyblocking_couples(self):
        blocking = KeyBlocking(
            ref_attr_index=1, target_attr_index=1, callback=partial(soundexcode, language="english")
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        pairs = list(blocking.iter_id_pairs())
        self.assertEqual(len(pairs), 8)
        for pair in SOUNDEX_PAIRS:
            self.assertIn(pair, pairs)

    def test_soundex_blocks(self):
        blocking = SoundexBlocking(ref_attr_index=1, target_attr_index=1, language="english")
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn((["a1", "a7"], ["b3", "b6"]), blocks)
        self.assertIn((["a2", "a5"], ["b4"]), blocks)
        self.assertIn((["a3"], ["b1", "b2"]), blocks)

    def test_soundex_couples(self):
        blocking = SoundexBlocking(ref_attr_index=1, target_attr_index=1, language="english")
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        pairs = list(blocking.iter_id_pairs())
        self.assertEqual(len(pairs), 8)
        for pair in SOUNDEX_PAIRS:
            self.assertIn(pair, pairs)


class NGramBlockingTest(unittest.TestCase):
    def test_ngram_blocks(self):
        blocking = NGramBlocking(ref_attr_index=1, target_attr_index=1)
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), 3)
        self.assertIn((["a3"], ["b1", "b2"]), blocks)
        self.assertIn((["a5"], ["b4"]), blocks)
        self.assertIn((["a1", "a4"], ["b3"]), blocks)

    def test_ngram_blocks_depth(self):
        blocking = NGramBlocking(ref_attr_index=1, target_attr_index=1, depth=1)
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), 5)
        self.assertIn((["a3"], ["b1", "b2"]), blocks)
        self.assertIn((["a5"], ["b4"]), blocks)
        self.assertIn((["a6"], ["b5"]), blocks)
        self.assertIn((["a7"], ["b6"]), blocks)
        self.assertIn((["a1", "a4"], ["b3"]), blocks)

    def test_ngram_blocks_2(self):
        refset = [["3", "ccdd", "aabb"], ["4", "ccdd", "bbaa"]]
        targetset = [["c", "ccdd", "aabb"], ["d", "ccdd", "bbaa"]]
        true_pairs = [("3", "c"), ("4", "d")]
        blocking = NGramBlocking(ref_attr_index=2, target_attr_index=2, ngram_size=2, depth=1)
        blocking.fit(refset, targetset)
        pairs = list(blocking.iter_id_pairs())
        self.assertEqual(len(pairs), len(true_pairs))


class SortedNeighborhoodBlockingTest(unittest.TestCase):
    def test_sorted_neighborhood_blocks(self):
        blocking = SortedNeighborhoodBlocking(ref_attr_index=1, target_attr_index=1, window_width=1)
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        true_blocks = [
            (["a6"], ["b7", "b5"]),
            (["a3"], ["b5", "b1"]),
            (["a2"], ["b2"]),
            (["a5"], ["b4"]),
            (["a7"], ["b4", "b6"]),
            (["a1"], ["b6", "b3"]),
            (["a4"], ["b3"]),
        ]
        self.assertEqual(len(blocks), len(true_blocks))
        for block in true_blocks:
            self.assertIn(block, blocks)

    def test_sorted_neighborhood_keyfunc(self):
        """Test sort reversing values"""
        blocking = SortedNeighborhoodBlocking(
            ref_attr_index=1, target_attr_index=1, key_func=lambda x: x[::-1], window_width=1
        )
        blocking.fit(SOUNDEX_REFSET, SOUNDEX_TARGETSET)
        blocks = list(blocking.iter_id_blocks())
        true_blocks = [
            (["a1"], ["b3"]),
            (["a2"], ["b6"]),
            (["a5"], ["b4"]),
            (["a3"], ["b7", "b1"]),
            (["a6"], ["b2", "b5"]),
            (["a4"], ["b5"]),
        ]
        self.assertEqual(len(blocks), len(true_blocks))
        for block in true_blocks:
            self.assertIn(block, blocks)


class MergeBlockingTest(unittest.TestCase):
    def test_merge_blocks(self):
        blocking = MergeBlocking(
            ref_attr_index=1, target_attr_index=None, score_func=lambda x: x[2]
        )
        refset = [
            ("http://fr.wikipedia.org/wiki/Paris_%28Texas%29", "Paris", 25898),
            ("http://fr.wikipedia.org/wiki/Paris", "Paris", 12223100),
            ("http://fr.wikipedia.org/wiki/Saint-Malo", "Saint-Malo", 46342),
        ]
        targetset = [("Paris (Texas)", 25000), ("Paris (France)", 12000000)]
        true_blocks = [
            (
                ["http://fr.wikipedia.org/wiki/Paris", "http://fr.wikipedia.org/wiki/Saint-Malo"],
                ["Paris (Texas)", "Paris (France)"],
            )
        ]
        blocking.fit(refset, targetset)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), len(true_blocks))
        self.assertEqual(len(blocks), len(true_blocks))
        for block in true_blocks:
            self.assertIn(block, blocks)

    def test_merge_blocks_targetset(self):
        blocking = MergeBlocking(
            ref_attr_index=None, target_attr_index=2, score_func=lambda x: x[1]
        )
        refset = [("Paris (Texas)", 25000), ("Paris (France)", 12000000)]
        targetset = [
            ("http://fr.wikipedia.org/wiki/Paris_%28Texas%29", 25898, "Paris"),
            ("http://fr.wikipedia.org/wiki/Paris", 12223100, "Paris"),
            ("http://fr.wikipedia.org/wiki/Saint-Malo", 46342, "Saint-Malo"),
        ]
        true_blocks = [
            (
                ["Paris (Texas)", "Paris (France)"],
                ["http://fr.wikipedia.org/wiki/Paris", "http://fr.wikipedia.org/wiki/Saint-Malo"],
            )
        ]
        blocking.fit(refset, targetset)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(len(blocks), len(true_blocks))
        self.assertEqual(len(blocks), len(true_blocks))
        for block in true_blocks:
            self.assertIn(block, blocks)


class KmeansBlockingTest(unittest.TestCase):
    def test_clustering_blocking_kmeans(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.2, 48.9)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        blocking = KmeansBlocking(ref_attr_index=2, target_attr_index=2)
        blocking.fit(refset, targetset)
        # Blocks
        blocks = list(blocking.iter_indice_blocks())
        self.assertEqual(len(blocks), 2)
        self.assertIn(([0, 1], [0, 2]), blocks)
        self.assertIn(([2, 3], [1]), blocks)
        # Pairs
        pairs = list(blocking.iter_indice_pairs())
        self.assertEqual(len(pairs), 6)
        for pair in ((0, 0), (0, 2), (1, 0), (1, 2), (2, 1), (3, 1)):
            self.assertIn(pair, pairs)


class MinHashingBlockingTest(unittest.TestCase):
    def test_minhashing(self):
        refset = [
            ["V1", "label1", "Un nuage flotta dans le grand ciel bleu."],
            ["V2", "label2", "Pour quelle occasion vous êtes-vous apprêtée ?"],
            ["V3", "label3", "Je les vis ensemble à plusieurs occasions."],
            ["V4", "label4", "Je n'aime pas ce genre de bandes dessinées tristes."],
            ["V5", "label5", "Ensemble et à plusieurs occasions, je les vis."],
        ]
        targetset = [
            ["T1", "labelt1", "Des grands nuages noirs flottent dans le ciel."],
            ["T2", "labelt2", "Je les ai vus ensemble à plusieurs occasions."],
            ["T3", "labelt3", "J'aime les bandes dessinées de genre comiques."],
        ]
        normalizer = SimplifyNormalizer(attr_index=2, lemmas=FRENCH_LEMMAS)
        refset = normalizer.normalize_dataset(refset)
        targetset = normalizer.normalize_dataset(targetset)
        blocking = MinHashingBlocking(threshold=0.4, ref_attr_index=2, target_attr_index=2)
        blocking.fit(refset, targetset)
        blocks = list(blocking.iter_indice_blocks())
        for align in (([2, 4], [1]), ([0], [0]), ([3], [2])):
            self.assertIn(align, blocks)


class KdTreeBlockingTest(unittest.TestCase):
    def test_kdtree(self):
        refset = [
            ["V1", "label1", (6.14194444444, 48.67)],
            ["V2", "label2", (6.2, 49)],
            ["V3", "label3", (5.1, 48)],
            ["V4", "label4", (5.2, 48.1)],
        ]
        targetset = [
            ["T1", "labelt1", (6.2, 48.9)],
            ["T2", "labelt2", (5.3, 48.2)],
            ["T3", "labelt3", (6.25, 48.91)],
        ]
        blocking = KdTreeBlocking(threshold=0.3, ref_attr_index=2, target_attr_index=2)
        blocking.fit(refset, targetset)
        blocks = list(blocking.iter_id_blocks())
        self.assertEqual(
            [(["V1"], ["T1", "T3"]), (["V2"], ["T1", "T3"]), (["V3"], ["T2"]), (["V4"], ["T2"])],
            blocks,
        )


class PipelineBlockingTest(unittest.TestCase):
    def test_pipeline_blocking(self):
        refset = [
            ["1", "aabb", "ccdd"],
            ["2", "aabb", "ddcc"],
            ["3", "ccdd", "aabb"],
            ["4", "ccdd", "bbaa"],
        ]
        targetset = [
            ["a", "aabb", "ccdd"],
            ["b", "aabb", "ddcc"],
            ["c", "ccdd", "aabb"],
            ["d", "ccdd", "bbaa"],
        ]
        true_pairs = [
            ((0, "1"), (0, "a")),
            ((1, "2"), (1, "b")),
            ((2, "3"), (2, "c")),
            ((3, "4"), (3, "d")),
        ]
        blocking_1 = NGramBlocking(ref_attr_index=1, target_attr_index=1, ngram_size=2, depth=1)
        blocking_2 = NGramBlocking(ref_attr_index=2, target_attr_index=2, ngram_size=2, depth=1)
        blocking = PipelineBlocking((blocking_1, blocking_2))
        blocking.fit(refset, targetset)
        pairs = list(blocking.iter_pairs())
        self.assertEqual(len(pairs), len(true_pairs))
        for pair in true_pairs:
            self.assertIn(pair, pairs)

    def test_pipeline_id_blocking(self):
        refset = [
            ["1", "aabb", "ccdd"],
            ["2", "aabb", "ddcc"],
            ["3", "ccdd", "aabb"],
            ["4", "ccdd", "bbaa"],
        ]
        targetset = [
            ["a", "aabb", "ccdd"],
            ["b", "aabb", "ddcc"],
            ["c", "ccdd", "aabb"],
            ["d", "ccdd", "bbaa"],
        ]
        true_pairs = [("1", "a"), ("2", "b"), ("3", "c"), ("4", "d")]
        blocking_1 = NGramBlocking(ref_attr_index=1, target_attr_index=1, ngram_size=2, depth=1)
        blocking_2 = NGramBlocking(ref_attr_index=2, target_attr_index=2, ngram_size=2, depth=1)
        blocking = PipelineBlocking((blocking_1, blocking_2))
        blocking.fit(refset, targetset)
        pairs = list(blocking.iter_id_pairs())
        self.assertEqual(len(pairs), len(true_pairs))
        for pair in true_pairs:
            self.assertIn(pair, pairs)


if __name__ == "__main__":
    unittest.main()



""" Tokenizer for sentences/words segmentation.
"""

import collections
import re

try:
    from nltk.tokenize.punkt import PunktSentenceTokenizer
except ImportError:
    NLTK_AVAILABLE = False
else:
    NLTK_AVAILABLE = True

Token = collections.namedtuple("Token", ["word", "start", "end", "sentence"])
Sentence = collections.namedtuple("Sentence", ["indice", "start", "end"])


class RichStringTokenizer:
    """Tokenizer for Yams' RichString content.

    The tokenizer uses a variable-length sliding window, i.e. a sliding
    window yielding tokens of N words.
    """

    def __init__(self, text, token_min_size=1, token_max_size=3):
        """
        :token_min_size: minimum number of words required to be a valid token
        :token_max_size: minimum number of words required to be a valid token
        """
        self.text = text
        self.token_min_size = token_min_size
        self.token_max_size = token_max_size

    words_re = r"[\w@-]+"

    def iter_tokens(self, text):
        """Iterate tokens over a text"""
        # Compute sentences
        sentences = self.find_sentences(text)
        # Compute words
        words = [m for m in re.finditer(self.words_re, text, re.UNICODE)]
        indice = 0
        while indice < len(words):
            # Choose the current sentence of the first word
            current_sentence = [s for s in sentences if s.start <= words[indice].start()][-1]
            # Sliding windows over the different words for each sentence
            remaining = len(words) - indice
            for length in range(min(self.token_max_size, remaining), self.token_min_size - 1, -1):
                _words = words[indice : indice + length]
                if _words[-1].start() > current_sentence.end:
                    # The last word in not in the same sentence anymore, split
                    continue
                normalized_word = " ".join([w.group() for w in _words]).strip()
                yield Token(normalized_word, _words[0].start(), _words[-1].end(), current_sentence)
            indice += 1

    @staticmethod
    def find_sentences(text):
        """Find the sentences"""
        if not NLTK_AVAILABLE:
            raise RuntimeError("find_sentences requires NLTK to be installed")
        sentences = PunktSentenceTokenizer().span_tokenize(text)
        return [Sentence(ind, start, end) for ind, (start, end) in enumerate(sentences)]

    def load_text(self, text):
        """Load the text to be tokenized"""
        self.text = text

    def __iter__(self):
        """Iterator over the text given in the object instantiation"""
        yield from self.iter_tokens(self.text)



# flake8: noqa
# Countries list (ISO-3166)
COUNTRIES = {
    "##": "non renseign\xc3\xa9",
    "..": "non renseign\xc3\xa9",
    "aa": "aire g\xc3\xa9ographique ancienne",
    "ad": "Andorre",
    "ae": "\xc3\x89mirats arabes unis",
    "af": "Afghanistan",
    "ag": "Antigua-et-Barbuda",
    "ai": "Anguilla",
    "al": "Albanie",
    "am": "Arm\xc3\xa9nie",
    "an": "Antilles n\xc3\xa9erlandaises",
    "ao": "Angola",
    "aq": "Antarctique",
    "ar": "Argentine",
    "as": "Samoa am\xc3\xa9ricaines",
    "at": "Autriche",
    "au": "Australie",
    "aw": "Aruba",
    "ax": "Aland (\xc3\xaeles)",
    "az": "Azerba\xc3\xafdjan",
    "ba": "Bosnie-Herz\xc3\xa9govine",
    "bb": "Barbade",
    "bd": "Bangladesh",
    "be": "Belgique",
    "bf": "Burkina",
    "bg": "Bulgarie",
    "bh": "Bahre\xc3\xafn",
    "bi": "Burundi",
    "bj": "B\xc3\xa9nin",
    "bl": "Saint-Barth\xc3\xa9lemy",
    "bm": "Bermudes",
    "bn": "Brun\xc3\xa9i",
    "bo": "Bolivie",
    "bq": "Bonaire, Saint-Eustache et Saba,Saba###Saint-Eustache",
    "br": "Br\xc3\xa9sil",
    "bs": "Bahamas",
    "bt": "Bhoutan",
    "bv": "Bouvet (\xc3\xaele)",
    "bw": "Botswana",
    "by": "Bi\xc3\xa9lorussie,B\xc3\xa9larus",
    "bz": "Belize",
    "ca": "Canada",
    "cc": "Cocos (\xc3\xaeles),Keeling (\xc3\xaeles)",
    "cd": "Congo (R\xc3\xa9publique d\xc3\xa9mocratique),Za\xc3\xafre",
    "cf": "Centrafrique,R\xc3\xa9publique centrafricaine",
    "cg": "Congo,Congo (R\xc3\xa9publique)",
    "ch": "Suisse,Conf\xc3\xa9d\xc3\xa9ration helv\xc3\xa9tique",
    "ci": "C\xc3\xb4te d'Ivoire\n",
    "ck": "Cook (\xc3\xaeles)",
    "cl": "Chili",
    "cm": "Cameroun",
    "cn": "Chine,Chine (R\xc3\xa9publique populaire)",
    "co": "Colombie",
    "cr": "Costa Rica",
    "cs": "Serbie-et-Mont\xc3\xa9n\xc3\xa9gro",
    "cu": "Cuba",
    "cv": "Cap-Vert",
    "cw": "Cura\xc3\xa7ao",
    "cx": "Christmas (\xc3\xaele)",
    "cy": "Chypre",
    "cz": "R\xc3\xa9publique tch\xc3\xa8que,Tch\xc3\xa8que, R\xc3\xa9publique",
    "dd": "Allemagne (R\xc3\xa9publique d\xc3\xa9mocratique)",
    "de": "Allemagne,Allemagne (R\xc3\xa9publique f\xc3\xa9d\xc3\xa9rale)",
    "dj": "Djibouti",
    "dk": "Danemark",
    "dm": "Dominique",
    "do": "R\xc3\xa9publique dominicaine,Dominicaine, R\xc3\xa9publique",
    "dz": "Alg\xc3\xa9rie",
    "ec": "\xc3\x89quateur",
    "ee": "Estonie",
    "eg": "\xc3\x89gypte",
    "eh": "Sahara occidental",
    "er": "\xc3\x89rythr\xc3\xa9e",
    "es": "Espagne",
    "et": "\xc3\x89thiopie",
    "fi": "Finlande",
    "fj": "Fidji",
    "fk": "Malouines (\xc3\xaeles),Falkland (\xc3\xaeles)",
    "fm": "Micron\xc3\xa9sie,\xc3\x89tats f\xc3\xa9d\xc3\xa9r\xc3\xa9s de Micron\xc3\xa9sie",
    "fo": "F\xc3\xa9ro\xc3\xa9 (\xc3\xaeles)",
    "fr": "France",
    "ga": "Gabon",
    "gb": "Grande-Bretagne,Royaume-Uni",
    "gd": "Grenade",
    "ge": "G\xc3\xa9orgie",
    "gf": "Guyane fran\xc3\xa7aise",
    "gg": "Guernesey",
    "gh": "Ghana",
    "gi": "Gibraltar",
    "gl": "Groenland",
    "gm": "Gambie",
    "gn": "Guin\xc3\xa9e",
    "gp": "Guadeloupe",
    "gq": "Guin\xc3\xa9e \xc3\xa9quatoriale",
    "gr": "Gr\xc3\xa8ce",
    "gs": "G\xc3\xa9orgie du Sud et les \xc3\xaeles Sandwich du Sud",
    "gt": "Guatemala",
    "gu": "Guam",
    "gw": "Guin\xc3\xa9e-Bissau",
    "gy": "Guyana",
    "hk": "Hong Kong",
    "hm": "Heard (\xc3\xaele) et \xc3\xaeles McDonald",
    "hn": "Honduras",
    "hr": "Croatie",
    "ht": "Ha\xc3\xafti",
    "hu": "Hongrie",
    "id": "Indon\xc3\xa9sie",
    "ie": "Irlande",
    "ii": "intergouvernemental",
    "il": "Isra\xc3\xabl",
    "im": "\xc3\x8ele de Man,Man, \xc3\x8ele de",
    "in": "Inde",
    "io": "Territoire britannique de l'Oc\xc3\xa9an indien,Chagos (\xc3\xaeles)###Oc\xc3\xa9an indien, Territoire britannique de l'\n",
    "iq": "Irak",
    "ir": "Iran",
    "is": "Islande",
    "it": "Italie",
    "je": "Jersey",
    "jm": "Jama\xc3\xafque",
    "jo": "Jordanie",
    "jp": "Japon",
    "ke": "Kenya",
    "kg": "Kirghizistan",
    "kh": "Cambodge",
    "ki": "Kiribati",
    "km": "Comores",
    "kn": "Saint-Kitts-et-Nevis,Saint-Christophe-et-Nevis",
    "ko": "Kosovo",
    "kp": "Cor\xc3\xa9e (R\xc3\xa9publique populaire d\xc3\xa9mocratique),Cor\xc3\xa9e du Nord",
    "kr": "Cor\xc3\xa9e (R\xc3\xa9publique),Cor\xc3\xa9e du Sud",
    "kw": "Kowe\xc3\xaft",
    "ky": "Cayman,Ca\xc3\xafmanes, \xc3\x8eles###Ca\xc3\xafman (\xc3\xaeles)",
    "kz": "Kazakhstan",
    "la": "Laos",
    "lb": "Liban",
    "lc": "Sainte-Lucie",
    "li": "Liechtenstein",
    "lk": "Sri Lanka",
    "lr": "Liberia",
    "ls": "Lesotho",
    "lt": "Lituanie",
    "lu": "Luxembourg",
    "lv": "Lettonie",
    "ly": "Libye",
    "ma": "Maroc",
    "mc": "Monaco",
    "md": "Moldavie,Moldova, R\xc3\xa9publique de",
    "me": "Mont\xc3\xa9n\xc3\xa9gro",
    "mf": "Saint-Martin (partie fran\xc3\xa7aise)",
    "mg": "Madagascar",
    "mh": "Marshall (\xc3\xaeles)",
    "mk": "Mac\xc3\xa9doine (R\xc3\xa9publique)",
    "ml": "Mali",
    "mm": "Myanmar,Birmanie",
    "mn": "Mongolie",
    "mo": "Macao",
    "mp": "Mariannes du Nord (\xc3\xaeles)",
    "mq": "Martinique",
    "mr": "Mauritanie",
    "ms": "Montserrat",
    "mt": "Malte",
    "mu": "Maurice",
    "mv": "Maldives",
    "mw": "Malawi",
    "mx": "Mexique",
    "my": "Malaisie",
    "mz": "Mozambique",
    "na": "Namibie",
    "nc": "Nouvelle-Cal\xc3\xa9donie",
    "ne": "Niger",
    "nf": "Norfolk (\xc3\xaele)",
    "ng": "Nigeria",
    "ni": "Nicaragua",
    "nl": "Pays-Bas",
    "no": "Norv\xc3\xa8ge",
    "np": "N\xc3\xa9pal",
    "nr": "Nauru",
    "nu": "Niue",
    "nz": "Nouvelle-Z\xc3\xa9lande",
    "om": "Oman",
    "oo": "code non adapt\xc3\xa9",
    "pa": "Panama",
    "pe": "P\xc3\xa9rou",
    "pf": "Polyn\xc3\xa9sie fran\xc3\xa7aise",
    "pg": "Papouasie-Nouvelle-Guin\xc3\xa9e",
    "ph": "Philippines",
    "pk": "Pakistan",
    "pl": "Pologne",
    "pm": "Saint-Pierre-et-Miquelon",
    "pn": "Pitcairn",
    "pr": "Porto Rico",
    "ps": "Autorit\xc3\xa9 palestinienne,Palestine",
    "pt": "Portugal",
    "pw": "Palau,Palaos",
    "py": "Paraguay",
    "qa": "Qatar",
    "re": "R\xc3\xa9union",
    "ro": "Roumanie",
    "rs": "Serbie",
    "ru": "Russie (F\xc3\xa9d\xc3\xa9ration),Russie",
    "rw": "Rwanda",
    "sa": "Arabie saoudite",
    "sb": "Salomon (\xc3\xaeles)",
    "sc": "Seychelles",
    "sd": "Soudan",
    "se": "Su\xc3\xa8de",
    "sg": "Singapour",
    "sh": "Sainte-H\xc3\xa9l\xc3\xa8ne,Ascension (\xc3\xaele)###Tristan da Cunha (\xc3\xaele)",
    "si": "Slov\xc3\xa9nie",
    "sj": "Svalbard et \xc3\xaele Jan Mayen",
    "sk": "Slovaquie",
    "sl": "Sierra Leone",
    "sm": "Saint-Marin",
    "sn": "S\xc3\xa9n\xc3\xa9gal",
    "so": "Somalie",
    "sr": "Suriname",
    "ss": "Soudan du Sud,Sud Soudan",
    "st": "Sao Tom\xc3\xa9-et-Principe",
    "su": "URSS",
    "sv": "El Salvador,Salvador",
    "sx": "Saint-Martin (partie n\xc3\xa9erlandaise),Sint Maarten",
    "sy": "Syrie",
    "sz": "Swaziland",
    "tc": "Turks et Ca\xc3\xafques (\xc3\xaeles)",
    "td": "Tchad",
    "tf": "Terres australes fran\xc3\xa7aises",
    "tg": "Togo",
    "th": "Tha\xc3\xaflande",
    "tj": "Tadjikistan",
    "tk": "Tokelau",
    "tl": "Timor oriental",
    "tm": "Turkm\xc3\xa9nistan",
    "tn": "Tunisie",
    "to": "Tonga",
    "tr": "Turquie",
    "tt": "Trinit\xc3\xa9-et-Tobago",
    "tv": "Tuvalu",
    "tw": "Ta\xc3\xafwan,Chine (R\xc3\xa9publique)",
    "tz": "Tanzanie",
    "ua": "Ukraine",
    "ug": "Ouganda",
    "um": "\xc3\x8eles mineures \xc3\xa9loign\xc3\xa9es des \xc3\x89tats-Unis",
    "us": "\xc3\x89tats-Unis",
    "uy": "Uruguay",
    "uz": "Ouzb\xc3\xa9kistan",
    "va": "Vatican,Saint-Si\xc3\xa8ge",
    "vc": "Saint-Vincent-et-les Grenadines",
    "ve": "Venezuela",
    "vg": "\xc3\x8eles Vierges britanniques,Vierges (\xc3\xaeles) britanniques",
    "vi": "\xc3\x8eles Vierges am\xc3\xa9ricaines,Vierges (\xc3\xaeles) am\xc3\xa9ricaines",
    "vn": "Viet Nam",
    "vu": "Vanuatu",
    "wf": "Wallis et Futuna (\xc3\xaeles)",
    "ws": "Samoa,Samoa occidentales",
    "xc": "Tch\xc3\xa9coslovaquie",
    "xd": "Allemagne avant 1945",
    "xe": "Europe,Union europ\xc3\xa9enne",
    "xk": "Cor\xc3\xa9e avant 1948",
    "xn": "Pays-Bas avant 1830,Belgique avant 1830",
    "xx": "inconnu",
    "yd": "Y\xc3\xa9men (R\xc3\xa9publique d\xc3\xa9mocratique populaire),Sud Y\xc3\xa9men",
    "ye": "Y\xc3\xa9men",
    "yt": "Mayotte",
    "yu": "Yougoslavie",
    "yy": "ne s'applique pas\n",
    "za": "Afrique du Sud",
    "zm": "Zambie",
    "zw": "Zimbabwe",
    "zz": "multiple\n",
}


# REGIONS TO COUNTRIES MAPPING
REGIONS_TO_COUNTRIES = {
    "Abruzzes": "Italie",
    "Acha\xefe": "Gr\xe8ce",
    "Acre": "Br\xe9sil",
    "Afghanistan": "Afghanistan",
    "Afrique du Sud": "Afrique du Sud",
    "Aguascalientes": "Mexique",
    "Ain": "France",
    "Aisne": "France",
    "Alabama": "\xc9tats-Unis",
    "Alagoas": "Br\xe9sil",
    "Aland (\xeeles)": "Aland (\xeeles)",
    "Alaska": "\xc9tats-Unis",
    "Albanie": "Albanie",
    "Alberta": "Canada",
    "Alg\xe9rie": "Alg\xe9rie",
    "Allemagne": "Allemagne",
    "Allemagne (R\xe9publique d\xe9mocratique)": "Allemagne (R\xe9publique d\xe9mocratique)",
    "Allemagne avant 1945": "Allemagne avant 1945",
    "Allier": "France",
    "Alpes-Maritimes": "France",
    "Alpes-de-Haute-Provence": "France",
    "Alsace": "France",
    "Amapa": "Br\xe9sil",
    "Amazonas": "Br\xe9sil",
    "Andalousie": "Espagne",
    "Andorre": "Andorre",
    "Angola": "Angola",
    "Anguilla": "Anguilla",
    "Antarctique": "Antarctique",
    "Antigua-et-Barbuda": "Antigua-et-Barbuda",
    "Antilles n\xe9erlandaises": "Antilles n\xe9erlandaises",
    "Anvers": "Belgique",
    "Appenzell-Rhodes-Ext\xe9rieures": "Suisse",
    "Appenzell-Rhodes-Int\xe9rieures": "Suisse",
    "Aquitaine": "France",
    "Arabie saoudite": "Arabie saoudite",
    "Aragon": "Espagne",
    "Arcadie": "Gr\xe8ce",
    "Ardennes": "France",
    "Ard\xe8che": "France",
    "Argentine": "Argentine",
    "Argolide": "Gr\xe8ce",
    "Argovie": "Suisse",
    "Arizona": "\xc9tats-Unis",
    "Ari\xe8ge": "France",
    "Arkansas": "\xc9tats-Unis",
    "Arm\xe9nie": "Arm\xe9nie",
    "Aruba": "Aruba",
    "Asturies": "Espagne",
    "Ath\xe8nes et agglom\xe9ration": "Gr\xe8ce",
    "Attique": "Gr\xe8ce",
    "Aube": "France",
    "Aude": "France",
    "Australie": "Australie",
    "Australie-M\xe9ridionale": "Australie",
    "Australie-Occidentale": "Australie",
    "Autorit\xe9 palestinienne": "Autorit\xe9 palestinienne",
    "Autriche": "Autriche",
    "Auvergne": "France",
    "Aveyron": "France",
    "Azerba\xefdjan": "Azerba\xefdjan",
    "Bade-Wurtemberg": "Allemagne",
    "Bahamas": "Bahamas",
    "Bahia": "Br\xe9sil",
    "Bahre\xefn": "Bahre\xefn",
    "Baja California Norte": "Mexique",
    "Baja California Sur": "Mexique",
    "Bangladesh": "Bangladesh",
    "Barbade": "Barbade",
    "Bas-Rhin": "France",
    "Basilicate": "Italie",
    "Basse-Autriche": "Autriche",
    "Basse-Normandie": "France",
    "Basse-Saxe": "Allemagne",
    "Bavi\xe8re": "Allemagne",
    "Belgique": "Belgique",
    "Belize": "Belize",
    "Berlin": "Allemagne",
    "Bermudes": "Bermudes",
    "Berne": "Suisse",
    "Bhoutan": "Bhoutan",
    "Bi\xe9lorussie": "Bi\xe9lorussie",
    "Bolivie": "Bolivie",
    "Bonaire, Saint-Eustache et Saba": "Bonaire, Saint-Eustache et Saba",
    "Bosnie-Herz\xe9govine": "Bosnie-Herz\xe9govine",
    "Botswana": "Botswana",
    "Bouches-du-Rh\xf4ne": "France",
    "Bourgogne": "France",
    "Bouvet (\xeele)": "Bouvet (\xeele)",
    "Brabant": "Belgique",
    "Brabant flamand": "Belgique",
    "Brabant wallon": "Belgique",
    "Brabant-Septentrional": "Pays-Bas",
    "Brandebourg": "Allemagne",
    "Bretagne": "France",
    "Brun\xe9i": "Brun\xe9i",
    "Bruxelles": "Belgique",
    "Br\xe9sil": "Br\xe9sil",
    "Br\xeame": "Allemagne",
    "Buenos Aires": "Argentine",
    "Bulgarie": "Bulgarie",
    "Burgenland": "Autriche",
    "Burkina": "Burkina",
    "Burundi": "Burundi",
    "B\xe2le-Campagne": "Suisse",
    "B\xe2le-Ville": "Suisse",
    "B\xe9nin": "B\xe9nin",
    "B\xe9otie": "Gr\xe8ce",
    "Calabre": "Italie",
    "Californie": "\xc9tats-Unis",
    "Calvados": "France",
    "Cambodge": "Cambodge",
    "Cameroun": "Cameroun",
    "Campanie": "Italie",
    "Campeche": "Mexique",
    "Canada": "Canada",
    "Canaries": "Espagne",
    "Cantabrie": "Espagne",
    "Cantal": "France",
    "Cap-Vert": "Cap-Vert",
    "Capitale f\xe9d\xe9rale": "Argentine",
    "Carinthie": "Autriche",
    "Caroline du Nord": "\xc9tats-Unis",
    "Caroline du Sud": "\xc9tats-Unis",
    "Castille et L\xe9on": "Espagne",
    "Castille-la Manche": "Espagne",
    "Catalogne": "Espagne",
    "Catamarca": "Argentine",
    "Cayman": "Cayman",
    "Cear\xe1": "Br\xe9sil",
    "Centrafrique": "Centrafrique",
    "Centre": "France",
    "Ceuta": "Espagne",
    "Chaco": "Argentine",
    "Chalcidique": "Gr\xe8ce",
    "Champagne-Ardenne": "France",
    "Charente": "France",
    "Charente-Maritime": "France",
    "Cher": "France",
    "Chiapas": "Mexique",
    "Chihuahua": "Mexique",
    "Chili": "Chili",
    "Chine": "Chine",
    "Christmas (\xeele)": "Christmas (\xeele)",
    "Chubut": "Argentine",
    "Chypre": "Chypre",
    "Ch\xedos": "Gr\xe8ce",
    "Coahuila": "Mexique",
    "Cocos (\xeeles)": "Cocos (\xeeles)",
    "Colima": "Mexique",
    "Colombie": "Colombie",
    "Colombie britannique": "Canada",
    "Colorado": "\xc9tats-Unis",
    "Communaut\xe9 de Madrid": "Espagne",
    "Communaut\xe9 de Valence": "Espagne",
    "Comores": "Comores",
    "Congo": "Congo",
    "Congo (R\xe9publique d\xe9mocratique)": "Congo (R\xe9publique d\xe9mocratique)",
    "Connecticut": "\xc9tats-Unis",
    "Cook (\xeeles)": "Cook (\xeeles)",
    "Corfou": "Gr\xe8ce",
    "Corinthie": "Gr\xe8ce",
    "Corrientes": "Argentine",
    "Corr\xe8ze": "France",
    "Corse": "France",
    "Corse-du-Sud": "France",
    "Cor\xe9e (R\xe9publique populaire d\xe9mocratique)": "Cor\xe9e (R\xe9publique populaire d\xe9mocratique)",
    "Cor\xe9e (R\xe9publique)": "Cor\xe9e (R\xe9publique)",
    "Cor\xe9e avant 1948": "Cor\xe9e avant 1948",
    "Costa Rica": "Costa Rica",
    "Creuse": "France",
    "Croatie": "Croatie",
    "Cr\xe8te": "Gr\xe8ce",
    "Cuba": "Cuba",
    "Cura\xe7ao": "Cura\xe7ao",
    "Cyclades": "Gr\xe8ce",
    "C\xe9phalonie": "Gr\xe8ce",
    "C\xf3rdoba": "Argentine",
    "C\xf4te d'Ivoire": "C\xf4te d'Ivoire",
    "C\xf4te-d'Or": "France",
    "C\xf4tes-d'Armor": "France",
    "Dakota du Nord": "\xc9tats-Unis",
    "Dakota du Sud": "\xc9tats-Unis",
    "Danemark": "Danemark",
    "Delaware": "\xc9tats-Unis",
    "Deux-S\xe8vres": "France",
    "District de Columbia": "\xc9tats-Unis",
    "District f\xe9d\xe9ral": "Br\xe9sil",
    "Djibouti": "Djibouti",
    "Dod\xe9can\xe8se": "Gr\xe8ce",
    "Dominique": "Dominique",
    "Dordogne": "France",
    "Doubs": "France",
    "Drenthe": "Pays-Bas",
    "Dr\xe1ma": "Gr\xe8ce",
    "Dr\xf4me": "France",
    "Durango": "Mexique",
    "D\xe9pendance de Ross (Nouvelle-Z\xe9lande)": "Antarctique",
    "El Salvador": "El Salvador",
    "Entre-Rios": "Argentine",
    "Espagne": "Espagne",
    "Espirito Santo": "Br\xe9sil",
    "Essonne": "France",
    "Estonie": "Estonie",
    "Estr\xe9madure": "Espagne",
    "Eub\xe9e": "Gr\xe8ce",
    "Eure": "France",
    "Eure-et-Loir": "France",
    "Eurytanie": "Gr\xe8ce",
    "Fidji": "Fidji",
    "Finist\xe8re": "France",
    "Finlande": "Finlande",
    "Flandre occidentale": "Belgique",
    "Flandre orientale": "Belgique",
    "Floride": "\xc9tats-Unis",
    "Fl\xf3rina": "Gr\xe8ce",
    "Formosa": "Argentine",
    "France": "France",
    "Franche-Comt\xe9": "France",
    "Fribourg": "Suisse",
    "Frioul-V\xe9n\xe9tie-Julienne": "Italie",
    "Frise": "Pays-Bas",
    "F\xe9ro\xe9 (\xeeles)": "F\xe9ro\xe9 (\xeeles)",
    "Gabon": "Gabon",
    "Galice": "Espagne",
    "Gambie": "Gambie",
    "Gard": "France",
    "Gen\xe8ve": "Suisse",
    "Gers": "France",
    "Ghana": "Ghana",
    "Gibraltar": "Gibraltar",
    "Gironde": "France",
    "Glaris": "Suisse",
    "Goi\xe1s": "Br\xe9sil",
    "Grande-Bretagne": "Grande-Bretagne",
    "Grenade": "Grenade",
    "Greven\xe1": "Gr\xe8ce",
    "Grisons": "Suisse",
    "Groenland": "Groenland",
    "Groningue": "Pays-Bas",
    "Gr\xe8ce": "Gr\xe8ce",
    "Gr\xe8ce centrale": "Gr\xe8ce",
    "Gr\xe8ce occidentale": "Gr\xe8ce",
    "Guadeloupe": "Guadeloupe",
    "Guam": "Guam",
    "Guanajuato": "Mexique",
    "Guatemala": "Guatemala",
    "Gueldre": "Pays-Bas",
    "Guernesey": "Guernesey",
    "Guerrero": "Mexique",
    "Guin\xe9e": "Guin\xe9e",
    "Guin\xe9e \xe9quatoriale": "Guin\xe9e \xe9quatoriale",
    "Guin\xe9e-Bissau": "Guin\xe9e-Bissau",
    "Guyana": "Guyana",
    "Guyane fran\xe7aise": "Guyane fran\xe7aise",
    "G\xe9orgie": "\xc9tats-Unis",
    "G\xe9orgie du Sud et les \xeeles Sandwich du Sud": "G\xe9orgie du Sud et les \xeeles Sandwich du Sud",
    "Hainaut": "Belgique",
    "Hambourg": "Allemagne",
    "Haut-Rhin": "France",
    "Haute-Autriche": "Autriche",
    "Haute-Corse": "France",
    "Haute-Garonne": "France",
    "Haute-Loire": "France",
    "Haute-Marne": "France",
    "Haute-Normandie": "France",
    "Haute-Savoie": "France",
    "Haute-Sa\xf4ne": "France",
    "Haute-Vienne": "France",
    "Hautes-Alpes": "France",
    "Hautes-Pyr\xe9n\xe9es": "France",
    "Hauts-de-Seine": "France",
    "Hawaii": "\xc9tats-Unis",
    "Ha\xefti": "Ha\xefti",
    "Heard (\xeele) et \xeeles McDonald": "Heard (\xeele) et \xeeles McDonald",
    "Hesse": "Allemagne",
    "Hidalgo": "Mexique",
    "Hollande-M\xe9ridionale": "Pays-Bas",
    "Hollande-Septentrionale": "Pays-Bas",
    "Honduras": "Honduras",
    "Hong Kong": "Hong Kong",
    "Hongrie": "Hongrie",
    "H\xe9rault": "France",
    "Idaho": "\xc9tats-Unis",
    "Ille-et-Vilaine": "France",
    "Illinois": "\xc9tats-Unis",
    "Inde": "Inde",
    "Indiana": "\xc9tats-Unis",
    "Indon\xe9sie": "Indon\xe9sie",
    "Indre": "France",
    "Indre-et-Loire": "France",
    "Iowa": "\xc9tats-Unis",
    "Io\xe1nnina": "Gr\xe8ce",
    "Irak": "Irak",
    "Iran": "Iran",
    "Irlande": "Irlande",
    "Ir\xe1kleion": "Gr\xe8ce",
    "Islande": "Islande",
    "Isra\xebl": "Isra\xebl",
    "Is\xe8re": "France",
    "Italie": "Italie",
    "Jalisco": "Mexique",
    "Jama\xefque": "Jama\xefque",
    "Japon": "Japon",
    "Jersey": "Jersey",
    "Jordanie": "Jordanie",
    "Jujuy": "Argentine",
    "Jura": "France",
    "Kansas": "\xc9tats-Unis",
    "Kard\xedtsa": "Gr\xe8ce",
    "Kastori\xe1": "Gr\xe8ce",
    "Kav\xe1la": "Gr\xe8ce",
    "Kazakhstan": "Kazakhstan",
    "Kentucky": "\xc9tats-Unis",
    "Kenya": "Kenya",
    "Kilk\xeds": "Gr\xe8ce",
    "Kirghizistan": "Kirghizistan",
    "Kiribati": "Kiribati",
    "Kosovo": "Kosovo",
    "Kowe\xeft": "Kowe\xeft",
    "Koz\xe1ni": "Gr\xe8ce",
    "La Can\xe9e": "Gr\xe8ce",
    "Laconie": "Gr\xe8ce",
    "Landes": "France",
    "Languedoc-Roussillon": "France",
    "Laos": "Laos",
    "Las\xedthi": "Gr\xe8ce",
    "Latium": "Italie",
    "Le Pir\xe9e": "Gr\xe8ce",
    "Lesotho": "Lesotho",
    "Lettonie": "Lettonie",
    "Leucade": "Gr\xe8ce",
    "Liban": "Liban",
    "Liberia": "Liberia",
    "Libye": "Libye",
    "Liechtenstein": "Liechtenstein",
    "Ligurie": "Italie",
    "Limbourg": "Pays-Bas",
    "Limousin": "France",
    "Lituanie": "Lituanie",
    "Li\xe8ge": "Belgique",
    "Loir-et-Cher": "France",
    "Loire": "France",
    "Loire-Atlantique": "France",
    "Loiret": "France",
    "Lombardie": "Italie",
    "Lorraine": "France",
    "Lot": "France",
    "Lot-et-Garonne": "France",
    "Louisiane": "\xc9tats-Unis",
    "Loz\xe8re": "France",
    "Lucerne": "Suisse",
    "Luxembourg": "Belgique",
    "L\xe1risa": "Gr\xe8ce",
    "L\xe9svos": "Gr\xe8ce",
    "Macao": "Macao",
    "Mac\xe9doine (R\xe9publique)": "Mac\xe9doine (R\xe9publique)",
    "Mac\xe9doine centrale": "Gr\xe8ce",
    "Mac\xe9doine occidentale": "Gr\xe8ce",
    "Mac\xe9doine orientale et Thrace": "Gr\xe8ce",
    "Madagascar": "Madagascar",
    "Magn\xe9sie": "Gr\xe8ce",
    "Maine": "\xc9tats-Unis",
    "Maine-et-Loire": "France",
    "Malaisie": "Malaisie",
    "Malawi": "Malawi",
    "Maldives": "Maldives",
    "Mali": "Mali",
    "Malouines (\xeeles)": "Malouines (\xeeles)",
    "Malte": "Malte",
    "Manche": "France",
    "Manitoba": "Canada",
    "Maranh\xe3o": "Br\xe9sil",
    "Marches": "Italie",
    "Mariannes du Nord (\xeeles)": "Mariannes du Nord (\xeeles)",
    "Marne": "France",
    "Maroc": "Maroc",
    "Marshall (\xeeles)": "Marshall (\xeeles)",
    "Martinique": "Martinique",
    "Maryland": "\xc9tats-Unis",
    "Massachusetts": "\xc9tats-Unis",
    "Mato grosso": "Br\xe9sil",
    "Mato grosso do Sul": "Br\xe9sil",
    "Maurice": "Maurice",
    "Mauritanie": "Mauritanie",
    "Mayenne": "France",
    "Mayotte": "Mayotte",
    "Mecklembourg-Pom\xe9ranie ant\xe9rieure": "Allemagne",
    "Melilla": "Espagne",
    "Mendoza": "Argentine",
    "Mess\xe9nie": "Gr\xe8ce",
    "Meurthe-et-Moselle": "France",
    "Meuse": "France",
    "Mexico": "Mexique",
    "Mexique": "Mexique",
    "Michigan": "\xc9tats-Unis",
    "Michoac\xe1n": "Mexique",
    "Micron\xe9sie": "Micron\xe9sie",
    "Midi-Pyr\xe9n\xe9es": "France",
    "Minas Gerais": "Br\xe9sil",
    "Minnesota": "\xc9tats-Unis",
    "Misiones": "Argentine",
    "Mississippi": "\xc9tats-Unis",
    "Missouri": "\xc9tats-Unis",
    "Moldavie": "Moldavie",
    "Molise": "Italie",
    "Monaco": "Monaco",
    "Mongolie": "Mongolie",
    "Montana": "\xc9tats-Unis",
    "Montserrat": "Montserrat",
    "Mont\xe9n\xe9gro": "Mont\xe9n\xe9gro",
    "Morbihan": "France",
    "Morelos": "Mexique",
    "Moselle": "France",
    "Mozambique": "Mozambique",
    "Murcie": "Espagne",
    "Myanmar": "Myanmar",
    "Namibie": "Namibie",
    "Namur": "Belgique",
    "Nauru": "Nauru",
    "Navarre": "Espagne",
    "Nayarit": "Mexique",
    "Nebraska": "\xc9tats-Unis",
    "Neuch\xe2tel": "Suisse",
    "Neuqu\xe9n": "Argentine",
    "Nevada": "\xc9tats-Unis",
    "New Hampshire": "\xc9tats-Unis",
    "New Jersey": "\xc9tats-Unis",
    "New York": "\xc9tats-Unis",
    "Nicaragua": "Nicaragua",
    "Nidwald": "Suisse",
    "Niger": "Niger",
    "Nigeria": "Nigeria",
    "Niue": "Niue",
    "Ni\xe8vre": "France",
    "Nord": "France",
    "Nord-Pas-de-Calais": "France",
    "Norfolk (\xeele)": "Norfolk (\xeele)",
    "Norv\xe8ge": "Norv\xe8ge",
    "Nouveau Mexique": "\xc9tats-Unis",
    "Nouveau-Brunswick": "Canada",
    "Nouvelle-Cal\xe9donie": "Nouvelle-Cal\xe9donie",
    "Nouvelle-Galles-du-Sud": "Australie",
    "Nouvelle-Z\xe9lande": "Nouvelle-Z\xe9lande",
    "Nouvelle-\xc9cosse": "Canada",
    "Nuevo Le\xf3n": "Mexique",
    "N\xe9pal": "N\xe9pal",
    "Oaxaca": "Mexique",
    "Obwald": "Suisse",
    "Ohio": "\xc9tats-Unis",
    "Oise": "France",
    "Oklahoma": "\xc9tats-Unis",
    "Oman": "Oman",
    "Ombrie": "Italie",
    "Ontario": "Canada",
    "Oregon": "\xc9tats-Unis",
    "Orne": "France",
    "Ouganda": "Ouganda",
    "Ouzb\xe9kistan": "Ouzb\xe9kistan",
    "Overijssell": "Pays-Bas",
    "Pakistan": "Pakistan",
    "Palau": "Palau",
    "Pampa": "Argentine",
    "Panama": "Panama",
    "Papouasie-Nouvelle-Guin\xe9e": "Papouasie-Nouvelle-Guin\xe9e",
    "Paraguay": "Paraguay",
    "Paraiba": "Br\xe9sil",
    "Param\xe1": "Br\xe9sil",
    "Paris": "France",
    "Par\xe1": "Br\xe9sil",
    "Pas-de-Calais": "France",
    "Pays Basque": "Espagne",
    "Pays-Bas": "Pays-Bas",
    "Pays-Bas avant 1830": "Pays-Bas avant 1830",
    "Pays-de-la-Loire": "France",
    "Pennsylvanie": "\xc9tats-Unis",
    "Pernambouc": "Br\xe9sil",
    "Philippines": "Philippines",
    "Phocide": "Gr\xe8ce",
    "Phtiotide": "Gr\xe8ce",
    "Piau\xed": "Br\xe9sil",
    "Picardie": "France",
    "Pitcairn": "Pitcairn",
    "Pi\xe9mont": "Italie",
    "Pi\xe9rie": "Gr\xe8ce",
    "Poitou-Charentes": "France",
    "Pologne": "Pologne",
    "Polyn\xe9sie fran\xe7aise": "Polyn\xe9sie fran\xe7aise",
    "Porto Rico": "Porto Rico",
    "Portugal": "Portugal",
    "Pouilles": "Italie",
    "Provence-Alpes-C\xf4te d'Azur": "France",
    "Pr\xe9veza": "Gr\xe8ce",
    "Puebla": "Mexique",
    "Puy-de-D\xf4me": "France",
    "Pyr\xe9n\xe9es-Atlantiques": "France",
    "Pyr\xe9n\xe9es-Orientales": "France",
    "P\xe9lla": "Gr\xe8ce",
    "P\xe9loponn\xe8se": "Gr\xe8ce",
    "P\xe9rou": "P\xe9rou",
    "Qatar": "Qatar",
    "Queensland": "Australie",
    "Quer\xe9taro": "Mexique",
    "Quintana Roo": "Mexique",
    "Qu\xe9bec": "Canada",
    "Rhode Island": "\xc9tats-Unis",
    "Rhodope": "Gr\xe8ce",
    "Rh\xe9nanie-Palatinat": "Allemagne",
    "Rh\xe9nanie-du-Nord-Westphalie": "Allemagne",
    "Rh\xf4ne": "France",
    "Rh\xf4ne-Alpes": "France",
    "Rio Grande do Norte": "Br\xe9sil",
    "Rio Grande do Sul": "Br\xe9sil",
    "Rio Negro": "Argentine",
    "Rio de Janeiro": "Br\xe9sil",
    "Rioja": "Argentine",
    "Rond\xf4nia": "Br\xe9sil",
    "Roraima": "Br\xe9sil",
    "Roumanie": "Roumanie",
    "Royaume-Uni": "Grande-Bretagne",
    "Russie (F\xe9d\xe9ration)": "Russie (F\xe9d\xe9ration)",
    "Rwanda": "Rwanda",
    "R\xc3\xa9publique Tch\xc3\xa8que": "R\xc3\xa9publique tch\xc3\xa8que",
    "R\xe9publique dominicaine": "R\xe9publique dominicaine",
    "R\xe9publique tch\xe8que": "R\xe9publique tch\xe8que",
    "R\xe9thymnon": "Gr\xe8ce",
    "R\xe9union": "R\xe9union",
    "Sahara occidental": "Sahara occidental",
    "Saint-Barth\xe9lemy": "Saint-Barth\xe9lemy",
    "Saint-Gall": "Suisse",
    "Saint-Kitts-et-Nevis": "Saint-Kitts-et-Nevis",
    "Saint-Marin": "Saint-Marin",
    "Saint-Martin (partie fran\xe7aise)": "Saint-Martin (partie fran\xe7aise)",
    "Saint-Martin (partie n\xe9erlandaise)": "Saint-Martin (partie n\xe9erlandaise)",
    "Saint-Pierre-et-Miquelon": "Saint-Pierre-et-Miquelon",
    "Saint-Vincent-et-les Grenadines": "Saint-Vincent-et-les Grenadines",
    "Sainte-H\xe9l\xe8ne": "Sainte-H\xe9l\xe8ne",
    "Sainte-Lucie": "Sainte-Lucie",
    "Salomon (\xeeles)": "Salomon (\xeeles)",
    "Salta": "Argentine",
    "Salzbourg": "Autriche",
    "Samoa": "Samoa",
    "Samoa am\xe9ricaines": "Samoa am\xe9ricaines",
    "San Juan": "Argentine",
    "San Luis": "Argentine",
    "San Luis Potos\xed": "Mexique",
    "Santa Catarina": "Br\xe9sil",
    "Santa Cruz": "Argentine",
    "Santa Fe": "Argentine",
    "Santiago del Estero": "Argentine",
    "Sao Tom\xe9-et-Principe": "Sao Tom\xe9-et-Principe",
    "Sardaigne": "Italie",
    "Sarre": "Allemagne",
    "Sarthe": "France",
    "Saskatchewan": "Canada",
    "Savoie": "France",
    "Saxe": "Allemagne",
    "Saxe-Anhalt": "Allemagne",
    "Sa\xf4ne-et-Loire": "France",
    "Schaffhouse": "Suisse",
    "Schleswig-Holstein": "Allemagne",
    "Schwyz": "Suisse",
    "Seine-Maritime": "France",
    "Seine-Saint-Denis": "France",
    "Seine-et-Marne": "France",
    "Serbie": "Serbie",
    "Serbie-et-Mont\xe9n\xe9gro": "Serbie-et-Mont\xe9n\xe9gro",
    "Sergipe": "Br\xe9sil",
    "Seychelles": "Seychelles",
    "Sicile": "Italie",
    "Sierra Leone": "Sierra Leone",
    "Sinaloa": "Mexique",
    "Singapour": "Singapour",
    "Slovaquie": "Slovaquie",
    "Slov\xe9nie": "Slov\xe9nie",
    "Soleure": "Suisse",
    "Somalie": "Somalie",
    "Somme": "France",
    "Sonora": "Mexique",
    "Soudan": "Soudan",
    "Soudan du Sud": "Soudan du Sud",
    "Sri Lanka": "Sri Lanka",
    "Styrie": "Autriche",
    "Suisse": "Suisse",
    "Suriname": "Suriname",
    "Su\xe8de": "Su\xe8de",
    "Svalbard et \xeele Jan Mayen": "Svalbard et \xeele Jan Mayen",
    "Swaziland": "Swaziland",
    "Syrie": "Syrie",
    "S\xe1mos": "Gr\xe8ce",
    "S\xe3o Paulo": "Br\xe9sil",
    "S\xe9n\xe9gal": "S\xe9n\xe9gal",
    "S\xe9rrai": "Gr\xe8ce",
    "Tabasco": "Mexique",
    "Tadjikistan": "Tadjikistan",
    "Tamaulipas": "Mexique",
    "Tanzanie": "Tanzanie",
    "Tarn": "France",
    "Tarn-et-Garonne": "France",
    "Tasmanie": "Australie",
    "Ta\xefwan": "Ta\xefwan",
    "Tchad": "Tchad",
    "Tch\xe9coslovaquie": "Tch\xe9coslovaquie",
    "Tennessee": "\xc9tats-Unis",
    "Terre de Feu": "Argentine",
    "Terre de la Reine-Maud (Norv\xe8ge)": "Antarctique",
    "Terre-Neuve": "Canada",
    "Terres australes et antarctiques fran\xe7aises": "Antarctique",
    "Terres australes fran\xe7aises": "Terres australes fran\xe7aises",
    "Territoire antarctique australien": "Antarctique",
    "Territoire antarctique britannique": "Antarctique",
    "Territoire britannique de l'Oc\xe9an indien": "Territoire britannique de l'Oc\xe9an indien",
    "Territoire de la capitale australienne": "Australie",
    "Territoire du Nord": "Australie",
    "Territoire du Yukon": "Canada",
    "Territoire-de-Belfort": "France",
    "Territoires du Nord-Ouest": "Canada",
    "Tessin": "Suisse",
    "Texas": "\xc9tats-Unis",
    "Tha\xeflande": "Tha\xeflande",
    "Thesprotie": "Gr\xe8ce",
    "Thessalie": "Gr\xe8ce",
    "Thessalonique": "Gr\xe8ce",
    "Thurgovie": "Suisse",
    "Thuringe": "Allemagne",
    "Timor oriental": "Timor oriental",
    "Tlaxcala": "Mexique",
    "Togo": "Togo",
    "Tokelau": "Tokelau",
    "Tonga": "Tonga",
    "Toscane": "Italie",
    "Trentin-Haut-Adige": "Italie",
    "Trinit\xe9-et-Tobago": "Trinit\xe9-et-Tobago",
    "Tr\xedkala": "Gr\xe8ce",
    "Tucum\xe1n": "Argentine",
    "Tunisie": "Tunisie",
    "Turkm\xe9nistan": "Turkm\xe9nistan",
    "Turks et Ca\xefques (\xeeles)": "Turks et Ca\xefques (\xeeles)",
    "Turquie": "Turquie",
    "Tuvalu": "Tuvalu",
    "Tyrol": "Autriche",
    "URSS": "URSS",
    "US": "\xc9tats-Unis",
    "USA": "\xc9tats-Unis",
    "Ukraine": "Ukraine",
    "Uri": "Suisse",
    "Uruguay": "Uruguay",
    "Utah": "\xc9tats-Unis",
    "Utrecht": "Pays-Bas",
    "Val d'Aoste": "Italie",
    "Val-d'Oise": "France",
    "Val-de-Marne": "France",
    "Valais": "Suisse",
    "Vanuatu": "Vanuatu",
    "Var": "France",
    "Vatican": "Vatican",
    "Vaucluse": "France",
    "Vaud": "Suisse",
    "Vend\xe9e": "France",
    "Venezuela": "Venezuela",
    "Veracruz": "Mexique",
    "Vermont": "\xc9tats-Unis",
    "Victoria": "Australie",
    "Vienne": "Autriche",
    "Viet Nam": "Viet Nam",
    "Virginie": "\xc9tats-Unis",
    "Virginie occidentale": "\xc9tats-Unis",
    "Vorarlberg": "Autriche",
    "Vosges": "France",
    "V\xe9n\xe9tie": "Italie",
    "Wallis et Futuna (\xeeles)": "Wallis et Futuna (\xeeles)",
    "Washington": "\xc9tats-Unis",
    "Wisconsin": "\xc9tats-Unis",
    "Wyoming": "\xc9tats-Unis",
    "X\xe1nthi": "Gr\xe8ce",
    "Yonne": "France",
    "Yougoslavie": "Yougoslavie",
    "Yucat\xe1n": "Mexique",
    "Yvelines": "France",
    "Y\xe9men": "Y\xe9men",
    "Y\xe9men (R\xe9publique d\xe9mocratique populaire)": "Y\xe9men (R\xe9publique d\xe9mocratique populaire)",
    "Zacatecas": "Mexique",
    "Zambie": "Zambie",
    "Zimbabwe": "Zimbabwe",
    "Zoug": "Suisse",
    "Zurich": "Suisse",
    "Z\xe1kynthos": "Gr\xe8ce",
    "Z\xe9lande": "Pays-Bas",
    "aire g\xe9ographique ancienne": "aire g\xe9ographique ancienne",
    "code non adapt\xe9": "code non adapt\xe9",
    "inconnu": "inconnu",
    "intergouvernemental": "intergouvernemental",
    "multiple": "multiple",
    "ne s'applique pas": "ne s'applique pas",
    "non renseign\xe9": "non renseign\xe9",
    "\xc1rta": "Gr\xe8ce",
    "\xc9gypte": "\xc9gypte",
    "\xc9lide": "Gr\xe8ce",
    "\xc9mathie": "Gr\xe8ce",
    "\xc9milie-Romagne": "Italie",
    "\xc9mirats arabes unis": "\xc9mirats arabes unis",
    "\xc9pire": "Gr\xe8ce",
    "\xc9quateur": "\xc9quateur",
    "\xc9rythr\xe9e": "\xc9rythr\xe9e",
    "\xc9tats-Unis": "\xc9tats-Unis",
    "\xc9thiopie": "\xc9thiopie",
    "\xc9tolie-et-Acarnanie": "Gr\xe8ce",
    "\xc9vros": "Gr\xe8ce",
    "\xcele Pierre 1er (Norv\xe8ge)": "Antarctique",
    "\xcele de Man": "\xcele de Man",
    "\xcele du Prince-\xc9douard": "Canada",
    "\xcele-de-France": "France",
    "\xceles Bal\xe9ares": "Espagne",
    "\xceles Ioniennes": "Gr\xe8ce",
    "\xceles Vierges am\xe9ricaines": "\xceles Vierges am\xe9ricaines",
    "\xceles Vierges britanniques": "\xceles Vierges britanniques",
    "\xceles de la Mer \xc9g\xe9e m\xe9ridionale": "Gr\xe8ce",
    "\xceles de la Mer \xc9g\xe9e septentrionale": "Gr\xe8ce",
    "\xceles mineures \xe9loign\xe9es des \xc9tats-Unis": "\xceles mineures \xe9loign\xe9es des \xc9tats-Unis",
}



""" Sources for Named Entities Recognition.
"""

from nazca.utils.dataio import sparqlquery


###############################################################################
# NER SOURCE ##################################################################
###############################################################################
class AbstractNerSource:
    """High-level source for Named Entities Recognition"""

    def __init__(self, endpoint, query, name=None, use_cache=True, preprocessors=None):
        """Initialise the class."""
        self.endpoint = endpoint
        self.query = query
        self.name = name
        self.preprocessors = preprocessors or []
        self.use_cache = use_cache
        self._recognized_cache = {}

    def add_preprocessors(self, preprocessor):
        """Add a preprocessor"""
        self.preprocessors.append(preprocessor)

    def recognize_token(self, token):
        """Recognize a token"""
        # Applies source specific preprocessors
        for preprocessor in self.preprocessors:
            token = preprocessor(token)
            if not token:
                return []
        if self.use_cache and token.word in self._recognized_cache:
            return self._recognized_cache[token.word]
        uris = self.query_word(token.word) if token.word else []
        if self.use_cache:
            self._recognized_cache[token.word] = uris
        return uris

    def query_word(self, word):
        """Query a word for a Named Entities Recognition process"""
        raise NotImplementedError


class NerSourceLexicon(AbstractNerSource):
    """Source based on a (pre-computed) dictionnary of words (token, uri)"""

    def __init__(self, lexicon, name=None, use_cache=True, preprocessors=None):
        self.lexicon = lexicon
        self.name = name
        self.preprocessors = preprocessors or []
        self.use_cache = use_cache
        self._recognized_cache = {}

    def query_word(self, word):
        uri = self.lexicon.get(word)
        return (
            [
                uri,
            ]
            if uri
            else []
        )


class NerSourceLocalRql(AbstractNerSource):
    """High-level source for Named Entities Recognition
    Local RQL version
    """

    def __init__(self, session, query, name=None, use_cache=True, preprocessors=None):
        """Initialise the class."""
        self.query = query
        self.session = session
        self.name = name
        self.preprocessors = preprocessors or []
        self.use_cache = use_cache
        self._recognized_cache = {}

    def query_word(self, word):
        """Query a word for a Named Entities Recognition process"""
        return [r[0] for r in self.session.execute(self.query, dict(word=word))]


class NerSourceSparql(AbstractNerSource):
    """High-level source for Named Entities Recognition
     SPARQL version

    >>> from ner.core import NerSourceSparql
    >>> ner_source = NerSourceSparql('''SELECT ?uri
                                          WHERE{
                                          ?uri rdfs:label "%(word)s"@en}''',
                                          'http://dbpedia.org/sparql')
    >>> print ner_source.recognize_token('Victor Hugo')
                 ... ['http://dbpedia.org/resource/Category:Victor_Hugo',
                      'http://dbpedia.org/resource/Victor_Hugo',
                      'http://dbpedia.org/class/yago/VictorHugo',
                      'http://dbpedia.org/class/yago/VictorHugo(ParisM%C3%A9tro)',
                      'http://sw.opencyc.org/2008/06/10/concept/en/VictorHugo',
                      'http://sw.opencyc.org/2008/06/10/concept/Mx4rve1ZXJwpEbGdrcN5Y29ycA']

    """

    def query_word(self, word):
        """Query a word for a Named Entities Recognition process"""
        return [r[0] for r in sparqlquery(self.endpoint, self.query % {"word": word})]



# pylint: disable-msg=W0622
# copyright 2004-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of nazca.
#
# nazca is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# nazca is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with nazca. If not, see <http://www.gnu.org/licenses/>.
"""Nazca packaging information."""
__docformat__ = "restructuredtext en"
from os.path import join
import sys

distname = "nazca"
modname = "nazca"

numversion = (3, 1, 0)
version = ".".join([str(num) for num in numversion])

license = "LGPL"  # 2.1 or later
description = "Python library for data alignment"
web = "https://forge.extranet.logilab.fr/open-source/nazca"
author = "Logilab"
author_email = "contact@logilab.fr"

classifiers = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Scientific/Engineering :: Information Analysis",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Text Processing",
]

__depends__ = {
    "python-dateutil": None,
    "lxml": ">= 4.0.0, < 6.0.0",
    "numpy": None,
    "scipy": None,
    "scikit-learn": None,
}

__recommends__ = {
    "sparqlwrapper": None,
    "nltk": None,
}

include_dirs = [join("test", "data"), "data", "examples", "ner", "rl", "utils"]

if sys.version_info < (2, 7):
    __depends__["unittest2"] = ">= 0.5.1"



#!/usr/bin/env python
# pylint: disable=W0404,W0622,W0704,W0613
# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Generic Setup script, takes package info from __pkginfo__.py file.
"""

__docformat__ = "restructuredtext en"

from os.path import join, dirname

from setuptools import find_packages, setup

here = dirname(__file__)

# load metadata from the __pkginfo__.py file so there is no risk of conflict
# see https://packaging.python.org/en/latest/single_source_version.html
pkginfo = join(here, "nazca", "__pkginfo__.py")
__pkginfo__ = {}
with open(pkginfo) as f:
    exec(f.read(), __pkginfo__)
# get required metadatas
distname = __pkginfo__["distname"]
version = __pkginfo__["version"]
license = __pkginfo__["license"]
description = __pkginfo__["description"]
web = __pkginfo__["web"]
author = __pkginfo__["author"]
author_email = __pkginfo__["author_email"]
classifiers = __pkginfo__["classifiers"]

with open(join(here, "README.rst")) as f:
    long_description = f.read()

# get optional metadatas
data_files = __pkginfo__.get("data_files", None)
dependency_links = __pkginfo__.get("dependency_links", ())

requires = {}
for entry in ("__depends__",):
    requires.update(__pkginfo__.get(entry, {}))
install_requires = ["{} {}".format(d, v and v or "").strip() for d, v in requires.items()]

setup(
    name=distname,
    version=version,
    license=license,
    description=description,
    long_description=long_description,
    author=author,
    author_email=author_email,
    url=web,
    classifiers=classifiers,
    packages=find_packages(exclude=["test", "examples"]),
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
    dependency_links=dependency_links,
)



#
# copyright 2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

import unittest

from nazca.utils import tokenizer
from nazca.ner import preprocessors


class PreprocessorTest(unittest.TestCase):
    """Test of preprocessors"""

    def test_lowercasefilter(self):
        preprocessor = preprocessors.NerLowerCaseFilterPreprocessor()
        token = tokenizer.Token("toto", 0, 4, None)
        self.assertEqual(preprocessor(token), None)
        token = tokenizer.Token("toto Tata", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        token = tokenizer.Token("toto tata", 0, 4, None)
        self.assertEqual(preprocessor(token), None)

    def test_wordsizefilter(self):
        preprocessor = preprocessors.NerWordSizeFilterPreprocessor()
        token = tokenizer.Token("toto", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        preprocessor = preprocessors.NerWordSizeFilterPreprocessor(min_size=3)
        token = tokenizer.Token("toto", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        token = tokenizer.Token("to", 0, 4, None)
        self.assertEqual(preprocessor(token), None)
        preprocessor = preprocessors.NerWordSizeFilterPreprocessor(max_size=3)
        token = tokenizer.Token("toto", 0, 4, None)
        self.assertEqual(preprocessor(token), None)
        token = tokenizer.Token("to", 0, 4, None)
        self.assertEqual(preprocessor(token), token)

    def test_lowerfirstword(self):
        preprocessor = preprocessors.NerLowerFirstWordPreprocessor()
        sentence = tokenizer.Sentence(0, 0, 20)
        # Start of the sentence
        token1 = tokenizer.Token("Toto tata", 0, 4, sentence)
        token2 = tokenizer.Token("Toto tata", 0, 4, sentence)
        self.assertEqual(preprocessor(token1), token2)
        token1 = tokenizer.Token("Us tata", 0, 4, sentence)
        token2 = tokenizer.Token("us tata", 0, 4, sentence)
        self.assertEqual(preprocessor(token1), token2)
        # Not start of the sentence
        token1 = tokenizer.Token("Toto tata", 12, 16, sentence)
        token2 = tokenizer.Token("Toto tata", 12, 16, sentence)
        self.assertEqual(preprocessor(token1), token2)
        token1 = tokenizer.Token("Us tata", 12, 16, sentence)
        token2 = tokenizer.Token("Us tata", 12, 16, sentence)
        self.assertEqual(preprocessor(token1), token2)

    def test_stopwordsfilter(self):
        preprocessor = preprocessors.NerStopwordsFilterPreprocessor()
        token = tokenizer.Token("Toto", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        token = tokenizer.Token("Us", 0, 4, None)
        self.assertEqual(preprocessor(token), None)
        token = tokenizer.Token("Us there", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        # Split words
        preprocessor = preprocessors.NerStopwordsFilterPreprocessor(split_words=True)
        token = tokenizer.Token("Us there", 0, 4, None)
        self.assertEqual(preprocessor(token), None)
        token = tokenizer.Token("Us there toto", 0, 4, None)
        self.assertEqual(preprocessor(token), token)

    def test_hashtag(self):
        preprocessor = preprocessors.NerHashTagPreprocessor()
        token = tokenizer.Token("Toto", 0, 4, None)
        self.assertEqual(preprocessor(token), token)
        token1 = tokenizer.Token("@BarackObama", 0, 4, None)
        token2 = tokenizer.Token("BarackObama", 0, 4, None)
        self.assertEqual(preprocessor(token1), token2)
        token1 = tokenizer.Token("@Barack_Obama", 0, 4, None)
        token2 = tokenizer.Token("Barack Obama", 0, 4, None)
        self.assertEqual(preprocessor(token1), token2)


if __name__ == "__main__":
    unittest.main()



# copyright 2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
import time
import logging
from collections import defaultdict

from numpy import zeros
from scipy.sparse import lil_matrix

from nazca.utils.dataio import parsefile


###############################################################################
# UTILITY FUNCTIONS ###########################################################
###############################################################################
def iter_aligned_pairs(refset, targetset, global_mat, global_matched, unique=True):
    """Return the aligned pairs"""
    if unique:
        for refid in global_matched:
            bestid, _ = sorted(global_matched[refid], key=lambda x: x[1])[0]
            ref_record = refset[refid]
            target_record = targetset[bestid]
            distance = global_mat[refid, bestid] if global_mat is not None else None
            yield (ref_record[0], refid), (target_record[0], bestid), distance
    else:
        for refid in global_matched:
            for targetid, _ in global_matched[refid]:
                ref_record = refset[refid]
                target_record = targetset[targetid]
                distance = global_mat[refid, targetid] if global_mat is not None else None
                yield (ref_record[0], refid), (target_record[0], targetid), distance


###############################################################################
# BASE ALIGNER OBJECT #########################################################
###############################################################################
class BaseAligner:
    def __init__(self, threshold, processings, normalize_matrix=False):
        self.threshold = threshold
        self.processings = processings
        self.normalize_matrix = normalize_matrix
        self.ref_normalizer = None
        self.target_normalizer = None
        self.target_normalizer = None
        self.blocking = None
        self.alignments_done = 0
        self.pairs_found = 0
        self.nb_comparisons = 0
        self.nb_blocks = 0
        self.refset_size = None
        self.targetset_size = None
        self.time = None
        self.logger = logging.getLogger("nazca.aligner")

    def register_ref_normalizer(self, normalizer):
        """Register normalizers to be applied
        before alignment"""
        self.ref_normalizer = normalizer

    def register_target_normalizer(self, normalizer):
        """Register normalizers to be applied
        before alignment"""
        self.target_normalizer = normalizer

    def register_blocking(self, blocking):
        self.blocking = blocking

    def apply_normalization(self, dataset, normalizer):
        if normalizer:
            return normalizer.normalize_dataset(dataset)
        return dataset

    def compute_distance_matrix(self, refset, targetset, ref_indexes, target_indexes):
        """Compute and return the global alignment matrix.
        For each `processing` a `Distancematrix` is built, then all the
        matrices are summed with their own weighting and the result is the global
        alignment matrix, which is returned.
        """
        distmatrix = zeros((len(ref_indexes), len(target_indexes)), dtype="float32")
        for processing in self.processings:
            w = processing.weight
            distmatrix += w * processing.cdist(refset, targetset, ref_indexes, target_indexes)
        return distmatrix

    def threshold_matched(self, distmatrix):
        """Return the matched elements within a dictionnary,
        each key being the indice from X, and the corresponding
        values being a list of couple (indice from Y, distance)
        """
        match = defaultdict(list)
        if self.normalize_matrix:
            distmatrix /= distmatrix.max()
        ind = (distmatrix <= self.threshold).nonzero()
        indrow = ind[0].tolist()
        indcol = ind[1].tolist()
        for i, j in zip(indrow, indcol):
            match[i].append((j, distmatrix[i, j]))
        return match

    def _get_match(self, refset, targetset, ref_indexes=None, target_indexes=None):
        # Build items
        ref_indexes = ref_indexes or range(len(refset))
        target_indexes = target_indexes or range(len(targetset))
        # Apply alignments
        mat = self.compute_distance_matrix(
            refset, targetset, ref_indexes=ref_indexes, target_indexes=target_indexes
        )
        matched = self.threshold_matched(mat)
        # Reapply matched to global indexes
        new_matched = {}
        for k, values in matched.items():
            new_matched[ref_indexes[k]] = [(target_indexes[i], d) for i, d in values]
        return mat, new_matched

    def align(self, refset, targetset, get_matrix=True):
        """Perform the alignment on the referenceset
        and the targetset
        """
        start_time = time.time()
        refset = self.apply_normalization(refset, self.ref_normalizer)
        targetset = self.apply_normalization(targetset, self.target_normalizer)
        self.refset_size = len(refset)
        self.targetset_size = len(targetset)
        # If no blocking
        if not self.blocking:
            return self._get_match(refset, targetset)
        # Blocking == conquer_and_divide
        global_matched = {}
        global_mat = lil_matrix((len(refset), len(targetset)))
        self.blocking.fit(refset, targetset)
        for refblock, targetblock in self.blocking.iter_blocks():
            self.nb_blocks += 1
            ref_index = [r[0] for r in refblock]
            target_index = [r[0] for r in targetblock]
            self.nb_comparisons += len(ref_index) * len(target_index)
            _, matched = self._get_match(refset, targetset, ref_index, target_index)
            for k, values in matched.items():
                subdict = global_matched.setdefault(k, set())
                for v, d in values:
                    subdict.add((v, d))
                    self.alignments_done += 1
                    if get_matrix:
                        # XXX avoid issue in sparse matrix
                        global_mat[k, v] = d or 10 ** (-10)
        self.time = time.time() - start_time
        return global_mat, global_matched

    def get_aligned_pairs(self, refset, targetset, unique=True, use_distance=True):
        """Get the pairs of aligned elements"""
        if not refset or not targetset:
            return
        global_mat, global_matched = self.align(refset, targetset, get_matrix=use_distance)
        for refitem, targetitem, distance in iter_aligned_pairs(
            refset, targetset, global_mat, global_matched, unique
        ):
            self.pairs_found += 1
            yield refitem, targetitem, distance
        self.log_infos()

    def align_from_files(
        self,
        reffile,
        targetfile,
        ref_indexes=None,
        target_indexes=None,
        ref_encoding=None,
        target_encoding=None,
        ref_separator="\t",
        target_separator="\t",
        get_matrix=True,
    ):
        """Align data from files

        Parameters
        ----------

        reffile: name of the reference file

        targetfile: name of the target file

        ref_encoding: if given (e.g. 'utf-8' or 'latin-1'), it will
                      be used to read the files.

        target_encoding: if given (e.g. 'utf-8' or 'latin-1'), it will
                         be used to read the files.

        ref_separator: separator of the reference file

        target_separator: separator of the target file
        """
        refset = parsefile(
            reffile, indexes=ref_indexes, encoding=ref_encoding, delimiter=ref_separator
        )
        targetset = parsefile(
            targetfile, indexes=target_indexes, encoding=target_encoding, delimiter=target_separator
        )
        return self.align(refset, targetset, get_matrix=get_matrix)

    def get_aligned_pairs_from_files(
        self,
        reffile,
        targetfile,
        ref_indexes=None,
        target_indexes=None,
        ref_encoding=None,
        target_encoding=None,
        ref_separator="\t",
        target_separator="\t",
        unique=True,
    ):
        """Get the pairs of aligned elements"""
        refset = parsefile(
            reffile, indexes=ref_indexes, encoding=ref_encoding, delimiter=ref_separator
        )
        targetset = parsefile(
            targetfile, indexes=target_indexes, encoding=target_encoding, delimiter=target_separator
        )
        if not refset or not targetset:
            return
        global_mat, global_matched = self.align(refset, targetset, get_matrix=False)
        yield from iter_aligned_pairs(refset, targetset, global_mat, global_matched, unique)

    def log_infos(self):
        """Display some info on the aligner process"""
        self.logger.info("Computation time : %s" % self.time)
        self.logger.info("Size reference set : %s" % self.refset_size)
        self.logger.info("Size target set : %s" % self.targetset_size)
        self.logger.info("Comparisons maximum : %s" % (self.refset_size * self.targetset_size))
        self.logger.info("Comparisons done : %s" % self.nb_comparisons)
        self.logger.info("Number of blocks : %s" % self.nb_blocks)
        self.logger.info(
            "Blocking reduction : %i%%"
            % (100 - (100 * self.nb_comparisons / float(self.refset_size * self.targetset_size)))
        )
        if self.nb_blocks:
            self.logger.info(
                "Ratio comparisons/block : %.1f" % (float(self.nb_comparisons) / self.nb_blocks)
            )
        self.logger.info("Alignments done : %s" % self.alignments_done)
        self.logger.info("Pairs found : %s" % self.pairs_found)
        self.logger.info(
            "Ratio alignments done / reference set : %i%%"
            % (100 * self.alignments_done / float(self.refset_size))
        )
        self.logger.info(
            "Ratio pairs found / reference set : %i%%"
            % (100 * self.pairs_found / float(self.refset_size))
        )
        self.logger.info(
            "Ratio alignments done / target set : %i%%"
            % (100 * self.alignments_done / float(self.targetset_size))
        )
        self.logger.info(
            "Ratio pairs found / target set : %i%%"
            % (100 * self.pairs_found / float(self.targetset_size))
        )


###############################################################################
# PIPELINE ALIGNER OBJECT ####################################################
###############################################################################
class PipelineAligner:
    """This pipeline will perform iterative alignments, removing each time
    the aligned results from the previous aligner.
    """

    def __init__(self, aligners):
        self.aligners = aligners
        self.pairs = {}
        self.nb_comparisons = 0
        self.nb_blocks = 0
        self.alignments_done = 0
        self.pairs_found = 0
        self.refset_size = None
        self.targetset_size = None
        self.time = None
        self.logger = logging.getLogger("nazca.aligner")

    def get_aligned_pairs(self, refset, targetset, unique=True):
        """Get the pairs of aligned elements"""
        if not refset or not targetset:
            return
        start_time = time.time()
        ref_index = range(len(refset))
        target_index = range(len(targetset))
        self.refset_size = len(refset)
        self.targetset_size = len(targetset)
        seen_refset = set()
        # Iteration over aligners
        for ind_aligner, aligner in enumerate(self.aligners):
            # Perform alignment
            _refset = [refset[i] for i in ref_index]
            _targetset = [targetset[i] for i in target_index]
            for refitem, targetitem, distance in aligner.get_aligned_pairs(
                _refset, _targetset, unique
            ):
                self.pairs_found += 1
                reindexed_refitem = (refitem[0], ref_index[refitem[1]])
                reindexed_targetitem = (targetitem[0], target_index[targetitem[1]])
                yield reindexed_refitem, reindexed_targetitem, distance
                seen_refset.add(ref_index[refitem[1]])
            # Store stats
            self.nb_blocks += aligner.nb_blocks
            self.nb_comparisons += aligner.nb_comparisons
            # Update indexes if necessary
            # For now, we remove all the reference set that are already matched
            if ind_aligner < len(self.aligners) - 1:
                # There are other aligners after this one
                ref_index = [i for i in ref_index if i not in seen_refset]
        self.time = time.time() - start_time
        self.log_infos()

    def log_infos(self):
        """Display some info on the aligner process"""
        self.logger.info("Computation time : %s" % self.time)
        self.logger.info("Size reference set : %s" % self.refset_size)
        self.logger.info("Size target set : %s" % self.targetset_size)
        self.logger.info("Maximum comparisons : %s" % (self.refset_size * self.targetset_size))
        self.logger.info("Comparisons done : %s" % self.nb_comparisons)
        self.logger.info("Number of blocks : %s" % self.nb_blocks)
        self.logger.info(
            "Blocking reduction : %i%%"
            % (100 - (100 * self.nb_comparisons / float(self.refset_size * self.targetset_size)))
        )
        if self.nb_blocks:
            self.logger.info(
                "Ratio comparisons/block : %.1f" % (float(self.nb_comparisons) / self.nb_blocks)
            )
        self.logger.info("Alignments done : %s" % self.alignments_done)
        self.logger.info("Pairs found : %s" % self.pairs_found)
        self.logger.info(
            "Ratio alignments done / reference set : %i%%"
            % (100 * self.alignments_done / float(self.refset_size))
        )
        self.logger.info(
            "Ratio pairs found / reference set : %i%%"
            % (100 * self.pairs_found / float(self.refset_size))
        )
        self.logger.info(
            "Ratio alignments done / target set : %i%%"
            % (100 * self.alignments_done / float(self.targetset_size))
        )
        self.logger.info(
            "Ratio pairs found / target set : %i%%"
            % (100 * self.pairs_found / float(self.targetset_size))
        )



""" Preprocessors for Named Entities Recognition.
"""

from nazca.utils.tokenizer import Token
from nazca.data.stopwords import FRENCH_STOPWORDS, ENGLISH_STOPWORDS

STOPWORDS = {"fr": FRENCH_STOPWORDS, "en": ENGLISH_STOPWORDS}


###############################################################################
# NER PREPROCESSORS ###########################################################
###############################################################################
class AbstractNerPreprocessor:
    """Preprocessor"""

    def __call__(self, token):
        raise NotImplementedError


class NerWordSizeFilterPreprocessor(AbstractNerPreprocessor):
    """Remove token based on the size of the word"""

    def __init__(self, min_size=None, max_size=None):
        self.min_size = min_size
        self.max_size = max_size

    def __call__(self, token):
        if (self.min_size and len(token.word) < self.min_size) or (
            self.max_size and len(token.word) > self.max_size
        ):
            return None
        return token


class NerLowerCaseFilterPreprocessor(AbstractNerPreprocessor):
    """Remove token with word in lower case"""

    def __call__(self, token):
        return None if token.word.islower() else token


class NerLowerFirstWordPreprocessor(AbstractNerPreprocessor):
    """Lower the first word of each sentence if it is a stopword."""

    def __init__(self, lang="en"):
        self.lang = lang

    def __call__(self, token):
        if token.start == token.sentence.start and token.word.split()[0].lower() in STOPWORDS.get(
            self.lang, ENGLISH_STOPWORDS
        ):
            word = token.word[0].lower() + token.word[1:]
            return Token(word, token.start, token.end, token.sentence)
        return token


class NerStopwordsFilterPreprocessor(AbstractNerPreprocessor):
    """Remove stopwords"""

    def __init__(self, split_words=False, lang="en"):
        self.split_words = split_words
        self.lang = lang

    def __call__(self, token):
        stopwords = STOPWORDS.get(self.lang, ENGLISH_STOPWORDS)
        if self.split_words and not [w for w in token.word.split() if w.lower() not in stopwords]:
            return None
        if not self.split_words and token.word.lower() in stopwords:
            return None
        return token


class NerHashTagPreprocessor(AbstractNerPreprocessor):
    """Cleanup hashtag"""

    def __call__(self, token):
        if token.word.startswith("@"):
            # XXX Split capitalize letter ?
            # @BarackObama -> Barack Obama
            word = token.word[1:].replace("_", " ")
            return Token(word, token.start, token.end, token.sentence)
        return token
